"""Test cases for the old scripts."""
from __future__ import annotations

import subprocess  # nosec B404
import sys
import typing
from pathlib import Path
from typing import Any
from typing import Iterator
from typing import List
from typing import Tuple

import matplotlib.testing.compare as mpltc  # type: ignore
import pytest
from matplotlib.testing.exceptions import ImageComparisonFailure  # type: ignore

_data = Path(__file__).parent / "data"
tmpoutput = _data / "_tmpoutput"
_expected = _data / "output"

Rscript = Tuple[Tuple[str, str, List[str]], Any]


@pytest.mark.skipif(sys.platform == "win32", reason="does not run on windows")
class TestTitrationFit:
    """Test the old ``fit_titration.py`` script."""

    note_fp = "./NTT-A04-Cl_note"
    csv_files = ["./Meas/A04 Cl_A.csv", "./Meas/A04 Cl_B.csv"]
    res_svd = [
        "K =  17.035\nsK =  0.666\nSA =  -0.0\nsSA =  0.007\nSB =  -0.275\nsSB =  0.002",
        "K =  14.824\nsK =  0.708\nSA =  0.019\nsSA =  0.004\nSB =  -0.274\nsSB =  0.002",
    ]
    res_band = [
        "K =  15.205\nsK =  0.549\nSA =  205506.748\nsSA =  1862.684\nSB =  1781.387\nsSB =  1249.163",
        "K =  15.015\nsK =  0.701\nSA =  253789.494\nsSA =  3055.602\nSB =  -4381.091\nsSB =  2040.904",
    ]

    @pytest.fixture(
        scope="class",
        params=[
            (csv_files[0], res_svd[0], ["-m", "svd"]),
            (csv_files[0], res_band[0], ["-m", "band", "-b", "480", "530"]),
            (csv_files[1], res_svd[1], ["-m", "svd"]),
            (csv_files[1], res_band[1], ["-m", "band", "-b", "480", "530"]),
        ],
    )
    def run_script(self, request: pytest.FixtureRequest) -> Iterator[Rscript]:
        """Run the script as class fixture."""
        cli = Path("../../src/clophfit/old/fit_titration.py")
        csv_file = request.param[0]
        with subprocess.Popen(  # nosec B603
            [cli, csv_file, self.note_fp, "-t", "cl", "-d", "_tmpoutput"]  # noqa: S603
            + request.param[2],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
            cwd=_data,
        ) as process:
            yield request.param, process.communicate()
            if (_data / "bs.txt").exists():
                (_data / "bs.txt").unlink()
            for fp in _expected.glob("*_pdf.png"):
                fp.unlink()

    def test_stdout(self, run_script: Rscript) -> None:
        """It print out results."""
        expected = run_script[0][1]
        assert expected in run_script[1][0]

    @pytest.mark.xfail(reason="Deprecation from dependency.")
    def test_stderr_svd(self, run_script: Rscript) -> None:
        """Test stderr for svd."""
        if run_script[0][2][1] == "svd":
            assert not run_script[1][1]
        else:
            raise AssertionError

    def test_stderr_band(self, run_script: Rscript) -> None:
        """Test stderr for band."""
        if run_script[0][2][1] == "band":
            assert not run_script[1][1]

    def test_pdf(self, run_script: Rscript) -> None:
        """It saves pdf file."""
        csv_ = run_script[0][0].split("/")[-1].split(".")[0]
        f = "_".join([run_script[0][2][1], csv_, "NTT-A04-Cl_note.pdf"])
        fp_test = tmpoutput / f
        fp_expected = _expected / f
        msg = mpltc.compare_images(fp_test, fp_expected, 80.0)
        if msg:
            raise ImageComparisonFailure(msg)  # pragma: no cover


@pytest.mark.skipif(sys.platform == "win32", reason="does not run on windows")
@typing.no_type_check
class TestTitrationFitGlobal:
    """It test the old ``fit_titration_global.py`` script."""

    dat_files = [
        "./global/pH/D05.dat",
        "./global/Cl/B05-20130628-cor.dat",
    ]
    res = [
        """SA1   683.357   714.804   740.747   767.245   800.925
SB1   246.164   299.394   338.212   374.611   417.296
SA2   11.0487   47.6023   76.2874     104.3   138.128
SB2    491.64   534.936   571.809   611.045   664.602
  K   7.34376   7.51152   7.65166   7.80003   7.99971
bootstrap:""",
        """SA1   26910.8   27072.2   27216.7   27361.3   27522.7
SB1   916.697   1037.35   1144.64    1251.2   1369.35
SA2   3813.97   3970.34   4110.42   4250.52   4406.96
SB2      -INF   24.2754   87.1125   149.801   219.626
  K   7.51099   7.78702   8.03881   8.29508   8.58663
bootstrap:""",
    ]

    @pytest.fixture(
        scope="class",
        params=[
            (dat_files[0], res[0], ["--boot", "3"]),
            (dat_files[1], res[1], ["--boot", "3", "-t", "cl"]),
        ],
    )
    def run_script(self, request: pytest.FixtureRequest) -> Iterator[Rscript]:
        """Run the script as class fixture."""
        cli = "../../src/clophfit/old/fit_titration_global.py"
        dat_file = request.param[0]
        with subprocess.Popen(  # nosec B603
            [cli, dat_file, "_tmpoutput"] + request.param[2],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
            cwd=_data,
            shell=False,  # noqa: S603
        ) as process:
            yield request.param, process.communicate()

    def test_stdout(self, run_script: Rscript) -> None:
        """It print out results."""
        expected = run_script[0][1]
        assert expected in run_script[1][0]

    @pytest.mark.xfail(reason="Deprecation from dependency.")
    def test_stderr(self, run_script: Rscript) -> None:
        """Stderr is empty."""
        assert not run_script[1][1]

    @pytest.mark.xfail(reason="Image sizes do not match.")
    def test_png(self, run_script: Rscript) -> None:
        """It saves pdf file."""
        f = ".".join([run_script[0][0], "png"])
        fp_test = tmpoutput / f.rsplit("/", maxsplit=1)[-1]
        fp_expected = _expected / f.lstrip("./")
        msg = mpltc.compare_images(fp_test, fp_expected, 1.01)
        if msg:  # pragma: no cover
            raise ImageComparisonFailure(msg)



"""Parse EnSpire files and optionally build titrations using a note file."""

from clophfit.prenspire.prenspire import CsvLineError
from clophfit.prenspire.prenspire import EnspireFile
from clophfit.prenspire.prenspire import ExpNote
from clophfit.prenspire.prenspire import ManyLinesFoundError
from clophfit.prenspire.prenspire import Titration
from clophfit.prenspire.prenspire import verbose_print

__all__ = [
    "CsvLineError",
    "EnspireFile",
    "ExpNote",
    "ManyLinesFoundError",
    "Titration",
    "verbose_print",
]



#!/usr/bin/python
#

import argparse
import pandas as pd
import matplotlib.pyplot as plt
# import seaborn as sns


def main():
    description = "plot a summary of a tecan fit analysis"
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('file',
                        help="file <well llower lower value upper uupper>")
    args = parser.parse_args()
    df = pd.read_table(args.file, sep=' ')
    # l = list(scheme.well[scheme.sample == 'buffer'])
    fig, ax = plt.subplots(figsize=(9, 12))
    # df.sort(columns=['lower'], inplace=True)
    N = len(df)
    ax.errorbar(df.pKa, range(N), xerr=[df.pKa-df.lower, df.upper-df.pKa], fmt='ok', ecolor='gray', alpha=0.5)
    ax.vlines([8.00], 0, N, linewidth=5, alpha=0.2)
    ax.set_xlabel("pH")
    ax.set_ylabel("mutant ID")
    ax.set_ylim([-1, N+1])
    # ax.set_xlim([4.5, 9.0])
    plt.show()


main()



"""Test cases for the binding functions module."""
from __future__ import annotations

import numpy as np
import pytest

from clophfit import binding
from clophfit.binding.fitting import fit_binding


def test_kd() -> None:
    """It returns int float np.ndarray[float]."""
    assert binding.kd(10, 7, 6.0) == 11
    assert binding.kd(10, 7, 7) == 20
    assert binding.kd(10, 7, 8) == 110.0
    np.testing.assert_allclose(
        binding.kd(10.0, 7.0, np.array([6.0, 8.0])), [11.0, 110.0]
    )
    np.testing.assert_allclose(binding.kd(10, 7, np.array([7, 6, 8.0])), [20, 11, 110])


def test_fit_titration() -> None:
    """It fits pH and Cl titrations."""
    x = [3.0, 5, 7, 9, 11.0]
    y = np.array([1.9991, 1.991, 1.5, 1.009, 1.0009])
    fit = binding.fit_titration("pH", x, y)
    assert abs(fit["K"][0] - 7) < 0.0000000001
    assert abs(fit["SA"][0] - 2) < 0.0001
    assert abs(fit["SB"][0] - 1) < 0.0001
    x = [0, 5.0, 10, 40, 160, 1000]
    y = np.array([2.0, 1.33333333, 1.0, 0.4, 0.11764706, 0.01980198])
    fit = binding.fit_titration("Cl", x, y)
    assert abs(fit["K"][0] - 10) < 0.0000001
    assert abs(fit["SA"][0] - 2) < 0.000000001
    assert abs(fit["SB"][0] - 0) < 0.00000001
    with pytest.raises(NameError, match="kind= pH or Cl"):
        binding.fit_titration("unk", x, y)


class TestFitBinding:
    """Test fit_binding()."""

    @staticmethod
    def ph_model(x: float, K: float, S0: float, S1: float) -> float:  # noqa: N803,D102
        return S0 + (S1 - S0) * 10 ** (K - x) / (1 + 10 ** (K - x))

    def test_fit_binding(self) -> None:
        """Fit binding curves."""
        x = np.array([9.0, 8.0, 7.0, 6.0, 5.0])
        y = np.array(
            [1.99, 1.909, 1.5, 1.0909, 1.0099]
        )  # This is generated by dummy_model with K=7, S0=2, S1=1
        result = fit_binding(x, y, self.ph_model)
        assert result.success is True
        assert np.isclose(result.params["K"].value, 7, 1e-4)
        assert np.isclose(result.params["S0"].value, 2, 1e-4)
        assert np.isclose(result.params["S1"].value, 1, 1e-5)

    def test_fit_binding_fail(self) -> None:
        """Raise ValueError."""
        x = np.array([1, 2, 3, 4, 5])
        y = np.array([2, 2, 2, 2, 1])  # This dataset doesn't fit the model
        with pytest.raises(ValueError, match="allowed"):
            fit_binding(x, y, self.ph_model)



"""Test prtecan module."""
from __future__ import annotations

import re
import sys
from pathlib import Path
from typing import Any

import numpy as np
import pandas as pd
import pytest

from clophfit import prtecan
from clophfit.prtecan import Labelblock
from clophfit.prtecan import TitrationAnalysis

# By defining csvl, lb0, and lb1 as class attributes, they are created only once
# per test session. Use fixture to capture UserWarning "OVER"

data_tests = Path(__file__).parent / "Tecan"
pytestmark = pytest.mark.filterwarnings("ignore:OVER")


def test_lookup_listoflines() -> None:
    """It returns indexes for pattern match on col in list of lines."""
    csvl: list[list[str | int | float]] = [
        ["pp", "xy", 1, 2.0],
        ["pp", "xx", 1, 2],
        ["pp", 12, 1, 2],
        ["pp", "yy", 1, 2.0],
        ["a"],
        ["pp", "xy", 1, 2],
    ]
    assert prtecan.lookup_listoflines(csvl, pattern="pp") == [0, 1, 2, 3, 5]
    assert not prtecan.lookup_listoflines(csvl, pattern="xy")  # == []
    assert prtecan.lookup_listoflines(csvl, pattern="xy", col=1) == [0, 5]
    assert prtecan.lookup_listoflines(csvl, pattern="yy", col=1) == [3]
    assert not prtecan.lookup_listoflines(csvl, pattern="yy", col=2)  # == []


def test_strip_lines() -> None:
    """It strips empty fields."""
    lines: list[list[float | int | str]] = [
        ["Excitation Wavelength", "", "", "", 485.0, "nm", "", "", ""]
    ]
    stripped = prtecan.strip_lines(lines)
    assert stripped == [["Excitation Wavelength", 485.0, "nm"]]


def test_extract_metadata() -> None:
    """It extracts metadata correctly."""
    lines: list[list[Any]] = [
        ["Label: Label1", "", "", "", "", "", "", "", "", "", "", "", ""],
        ["Mode", "", "", "", "Fluorescence Top Reading", "", "", "", "", ""],
        ["Shaking (Linear) Amplitude:", "", "", "", 2, "mm", "", "", "", "", ""],
        ["Excitation Wavelength", "", "", "", 400, "nm", "", "unexpected", "", "", ""],
        ["", "Temperature: 26 °C", "", "", "", "", "", "", "", "", ""],
    ]
    expected_metadata = {
        "Label": prtecan.Metadata("Label1"),
        "Mode": prtecan.Metadata("Fluorescence Top Reading"),
        "Shaking (Linear) Amplitude:": prtecan.Metadata(2, ["mm"]),
        "Excitation Wavelength": prtecan.Metadata(400, ["nm", "unexpected"]),
        "Temperature": prtecan.Metadata(26.0, ["°C"]),
    }

    metadata = prtecan.extract_metadata(lines)
    assert metadata == expected_metadata


def test_merge_md() -> None:
    """Merge metadata of both labelblocks and tecanfiles."""
    md1 = {
        "Gain": prtecan.Metadata(93, ["Manual"]),
        "Shaking (Linear) Amplitude:": prtecan.Metadata(2, ["mm"]),
    }
    md2 = {
        "Gain": prtecan.Metadata(93, ["Optimal"]),
        "Shaking (Linear) Amplitude:": prtecan.Metadata(2, ["mm"]),
    }
    mmd = prtecan.merge_md([md1, md2])
    assert mmd["Gain"] == prtecan.Metadata(93)
    assert mmd["Shaking (Linear) Amplitude:"] == prtecan.Metadata(2, ["mm"])


def test_calculate_conc() -> None:
    """Calculates concentration values from Cl additions."""
    additions = [112, 2, 2, 2, 2, 2, 2, 6, 4]
    conc = prtecan.calculate_conc(additions, 1000)
    np.testing.assert_almost_equal(
        conc,
        [0.0, 17.544, 34.483, 50.847, 66.667, 81.967, 96.774, 138.462, 164.179],
        3,
    )


@pytest.mark.filterwarnings("ignore:OVER")
class TestLabelblock:
    """Test labelblock class."""

    @staticmethod
    def _get_two_labelblocks() -> tuple[Labelblock, Labelblock]:
        """Simulate csvl with 2 labelblocks."""
        csvl = prtecan.read_xls(data_tests / "140220/pH6.5_200214.xls")
        idxs = prtecan.lookup_listoflines(csvl)
        # pylint: disable=W0201
        lb0 = Labelblock(csvl[idxs[0] : idxs[1]])
        lb1 = Labelblock(csvl[idxs[1] :])
        lb0.buffer_wells = ["D01", "D12", "E01", "E12"]
        lb1.buffer_wells = ["D01", "D12", "E01", "E12"]
        return lb0, lb1

    @pytest.fixture(scope="class")
    def labelblocks(self) -> tuple[Labelblock, Labelblock]:
        """Fixture that provides two labelblocks."""
        return self._get_two_labelblocks()

    def test_metadata(self, labelblocks: tuple[Labelblock, Labelblock]) -> None:
        """It parses "Temperature" metadata."""
        lb0, lb1 = labelblocks
        assert lb0.metadata["Temperature"].value == 25.6
        assert lb1.metadata["Temperature"].value == 25.3

    def test_data(self, labelblocks: tuple[Labelblock, Labelblock]) -> None:
        """It parses data values."""
        lb0, lb1 = labelblocks
        assert lb0.data["F06"] == 19551
        assert lb1.data["H12"] == 543

    def test_data_normalized(self, labelblocks: tuple[Labelblock, Labelblock]) -> None:
        """Normalize data using some metadata values."""
        lb0, lb1 = labelblocks
        assert lb0.data_norm["F06"] == pytest.approx(1051.1290323)
        assert lb1.data_norm["H12"] == pytest.approx(48.4821429)

    def test_data_buffersubtracted(
        self, labelblocks: tuple[Labelblock, Labelblock]
    ) -> None:
        """Calculate buffer value from average of buffer wells and subtract from data."""
        lb0, lb1 = labelblocks
        assert lb0.buffer == 11889.25
        assert lb1.buffer == 56.75
        assert lb0.buffer_sd == pytest.approx(450.2490)
        assert lb1.buffer_sd == pytest.approx(4.43706)
        assert lb0.data_buffersubtracted["F06"] == pytest.approx(7661.75)
        assert lb1.data_buffersubtracted["H12"] == pytest.approx(486.25)
        # Can also assign a buffer value.
        lb0.buffer = 1
        lb1.buffer = 2.9
        assert lb0.data_buffersubtracted["F06"] == 19550
        assert lb1.data_buffersubtracted["H12"] == 540.1

    def test_data_buffersubtracted_norm(
        self, labelblocks: tuple[Labelblock, Labelblock]
    ) -> None:
        """Calculate normalized buffer value from average of buffer wells and subtract from data."""
        lb0, lb1 = labelblocks
        assert lb0.buffer_norm == pytest.approx(639.20699)
        assert lb1.buffer_norm == pytest.approx(5.06696)
        assert lb0.buffer_norm_sd == pytest.approx(24.20694)
        assert lb1.buffer_norm_sd == pytest.approx(0.396166)
        assert lb0.data_buffersubtracted_norm["F06"] == pytest.approx(411.922)
        assert lb1.data_buffersubtracted_norm["H12"] == pytest.approx(43.4152)
        # Can also assign a buffer_norm value.
        lb0.buffer_norm = 1
        lb1.buffer_norm = 0.4821
        assert lb0.data_buffersubtracted_norm["F06"] == pytest.approx(1050.13)
        assert lb1.data_buffersubtracted_norm["H12"] == pytest.approx(48.0)

    def test_eq(self, labelblocks: tuple[Labelblock, Labelblock]) -> None:
        """Check if a Labelblock is equal to itself and not equal to a different Labelblock."""
        lb0, lb1 = labelblocks
        assert lb0 == lb0, "Labelblock is not equal to itself"
        assert lb0 != lb1, "Different Labelblocks are incorrectly reported as equal"
        assert (
            lb0.__eq__(1) == NotImplemented
        ), "Equality check against non-Labelblock object did not return NotImplemented"

    def test_almost_eq(self, labelblocks: tuple[Labelblock, Labelblock]) -> None:
        """Test the __almost_eq__ method of the Labelblock class."""
        lb0, _ = labelblocks
        file_path1 = Path(data_tests) / "290513_7.2.xls"
        csvl1 = prtecan.read_xls(file_path1)  # Gain=98
        idxs1 = prtecan.lookup_listoflines(csvl1)
        lb11 = Labelblock(csvl1[idxs1[1] :])
        file_path2 = Path(data_tests) / "290513_8.8.xls"
        csvl2 = prtecan.read_xls(file_path2)  # Gain=99
        idxs2 = prtecan.lookup_listoflines(csvl2)
        lb12 = Labelblock(csvl2[idxs2[1] :])
        assert lb11 != lb12
        assert lb11.__almost_eq__(lb12)
        assert not lb11.__almost_eq__(lb0)

    def test_overvalue(self) -> None:
        """It detects saturated data ("OVER")."""
        csvl = prtecan.read_xls(data_tests / "140220/pH6.5_200214.xls")
        idxs = prtecan.lookup_listoflines(csvl)
        with pytest.warns(
            UserWarning, match=r"OVER\n Overvalue in Label1:A06 of tecanfile "
        ):
            lb = Labelblock(csvl[idxs[0] : idxs[1]])
        assert np.nansum(lb.data["A06"]) == np.nansum(np.nan)
        assert np.nansum(lb.data["H02"]) == np.nansum(np.nan)

    def test_raise_missing_column(self) -> None:
        """It raises Exception when a column is missing from the labelblock."""
        csvl = prtecan.read_xls(data_tests / "exceptions/88wells_290212_20.xlsx")
        idxs = prtecan.lookup_listoflines(csvl)
        with pytest.raises(ValueError, match=r"Cannot build Labelblock: not 96 wells?"):
            Labelblock(csvl[idxs[0] : len(csvl)])

    def test_raise_missing_row(self) -> None:
        """It raises Exception when a row is missing from the labelblock."""
        csvl = prtecan.read_xls(data_tests / "exceptions/84wells_290212_20.xlsx")
        idxs = prtecan.lookup_listoflines(csvl)
        with pytest.raises(
            ValueError, match="Cannot extract data in Labelblock: not 96 wells?"
        ):
            Labelblock(csvl[idxs[0] : len(csvl)])


class TestCsvlFunctions:
    """Test TecanFile reading and parsing functions."""

    csvl = prtecan.read_xls(data_tests / "140220/pH8.3_200214.xls")

    def test_read_xls(self) -> None:
        """The test reads the xls file using cls method."""
        assert len(self.csvl) == 74

    def test_lookup_listoflines(self) -> None:
        """It finds Label occurrences using module function."""
        assert prtecan.lookup_listoflines(self.csvl) == [14, 44]


class TestTecanfile:
    """Test TecanFile class."""

    tf = prtecan.Tecanfile(data_tests / "140220/pH8.3_200214.xls")

    def test_path(self) -> None:
        """It reads the file path."""
        assert self.tf.path == data_tests / "140220/pH8.3_200214.xls"

    def test_metadata(self) -> None:
        """It parses the Date."""
        assert self.tf.metadata["Date:"].value == "20/02/2014"

    def test_labelblocks(self) -> None:
        """It parses "Temperature" metadata and cell data from 2 labelblocks."""
        assert self.tf.labelblocks[0].metadata["Temperature"].value == 25.3
        assert self.tf.labelblocks[1].metadata["Temperature"].value == 25.7
        assert self.tf.labelblocks[0].data["A01"] == 17260
        assert self.tf.labelblocks[1].data["H12"] == 4196

    def test_eq(self) -> None:
        """Check if a Tecanfile is equal to itself and not equal to a different Tecanfile."""
        tf1 = prtecan.Tecanfile(data_tests / "140220/pH8.3_200214.xls")
        assert self.tf == tf1, "Tecanfile is not equal to itself"
        tf2 = prtecan.Tecanfile(data_tests / "140220/pH9.1_200214.xls")
        assert self.tf != tf2, "Different Tecanfiles are incorrectly reported as equal"

    def test_warn(self) -> None:
        """It warns when labelblocks are repeated in a Tf as it might compromise grouping."""
        with pytest.warns(UserWarning, match="Repeated labelblocks"):
            prtecan.Tecanfile(data_tests / "exceptions/290212_7.67_repeated_lb.xls")

    def test_filenotfound(self) -> None:
        """It raises FileNotFoundError when the file path does not exist."""
        with pytest.raises(FileNotFoundError):
            prtecan.Tecanfile(Path("pinocchio"))

    def test_missing_label(self) -> None:
        """It raises Exception when there is no Label pattern."""
        with pytest.raises(ValueError, match="No Labelblock found."):
            prtecan.Tecanfile(data_tests / "exceptions/0_Labelblocks_290513_5.5.xlsx")


class TestLabelblocksGroup:
    """Test LabelBlocksGroup class."""

    tfs = [
        prtecan.Tecanfile(data_tests / "290513_5.5.xls"),
        prtecan.Tecanfile(data_tests / "290513_7.2.xls"),
        prtecan.Tecanfile(data_tests / "290513_8.8.xls"),
    ]
    lbg0 = prtecan.LabelblocksGroup([tfs[0].labelblocks[0], tfs[1].labelblocks[0]])
    lbg1 = prtecan.LabelblocksGroup([tfs[1].labelblocks[1], tfs[2].labelblocks[1]])
    lbg0.buffer_wells = ["C12", "D01", "D12", "E01", "E12", "F01"]
    lbg1.buffer_wells = ["C12", "D01", "D12", "E01", "E12", "F01"]

    def test_metadata(self) -> None:
        """Merge only shared metadata."""
        assert self.lbg0.metadata.get("Temperature") is None
        assert self.lbg1.metadata.get("Temperature") is None
        assert self.lbg1.metadata.get("Gain") is None
        assert self.lbg1.labelblocks[0].metadata["Gain"].value == 98
        assert self.lbg1.labelblocks[1].metadata["Gain"].value == 99
        # Common metadata.
        assert self.lbg0.metadata["Gain"].value == 94
        assert self.lbg0.metadata["Number of Flashes"].value == 10

    def test_data(self) -> None:
        """Merge data."""
        assert self.lbg0.data is not None
        assert self.lbg0.data["A01"] == [18713, 17088]
        assert self.lbg0.data["H12"] == [28596, 25771]
        assert self.lbg1.data is None

    def test_data_normalized(self) -> None:
        """Merge data_normalized."""
        np.testing.assert_almost_equal(
            self.lbg1.data_norm["H12"], [693.980, 714.495], 3
        )
        np.testing.assert_almost_equal(
            self.lbg0.data_norm["A01"], [995.372, 908.936], 3
        )

    def test_data_buffersubtracted(self) -> None:
        """Merge data_buffersubtracted."""
        assert self.lbg0.data_buffersubtracted is not None
        np.testing.assert_almost_equal(
            self.lbg0.data_buffersubtracted["B07"], [7069, 5716.7], 1
        )
        assert self.lbg1.data_buffersubtracted is None

    def test_data_buffersubtracted_norm(self) -> None:
        """Merge data_buffersubtracted."""
        np.testing.assert_almost_equal(
            self.lbg0.data_buffersubtracted_norm["B07"], [376.01, 304.08], 2
        )
        np.testing.assert_almost_equal(
            self.lbg1.data_buffersubtracted_norm["B07"], [355.16, 348.57], 2
        )

    def test_notequal_labelblocks(self) -> None:
        """It raises Exception when concatenating unequal labelblocks."""
        tfs = self.tfs
        with pytest.raises(ValueError, match="Creation of labelblock group failed."):
            prtecan.LabelblocksGroup([tfs[1].labelblocks[0], tfs[2].labelblocks[1]])


class TestTecanfileGroup:
    """Group tecanfiles properly."""

    class TestAllEqLbgs:
        """Test TfG with 2 LbG in the same order."""

        filenames = ["290513_5.5.xls", "290513_7.2.xls"]
        tecanfiles = [prtecan.Tecanfile(data_tests / f) for f in filenames]
        tfg = prtecan.TecanfilesGroup(tecanfiles)

        def test_metadata(self) -> None:
            """Parse general metadata."""
            assert (
                self.tfg.metadata["Plate"].value
                == "PE 96 Flat Bottom White   [PE.pdfx]"
            )
            assert self.tfg.metadata["System"].value == "TECANROBOT"

        def test_labelblocksgroups(self) -> None:
            """Generate 2 LbG with .data and .metadata."""
            lbg0 = self.tfg.labelblocksgroups[0]
            lbg1 = self.tfg.labelblocksgroups[1]
            # metadata
            assert lbg0.metadata["Number of Flashes"].value == 10.0
            assert lbg1.metadata["Gain"].value == 98.0
            # data normalized ... enough in lbg
            # data
            assert lbg0.data is not None
            assert lbg0.data["A01"] == [18713, 17088]
            assert lbg0.data["H12"] == [28596, 25771]
            assert lbg1.data is not None
            assert lbg1.data["A01"] == [7878, 8761]
            assert lbg1.data["H12"] == [14226, 13602]

    class TestAlmostEqLbgs:
        """Test TfG when 1 LbG equal and a second with almost equal labelblocks."""

        filenames = [
            "290513_5.5.xls",  # Label1 and Label2
            "290513_7.2.xls",  # Label1 and Label2
            "290513_8.8.xls",  # Label1 and Label2 with different metadata
        ]
        tecanfiles = [prtecan.Tecanfile(data_tests / f) for f in filenames]
        with pytest.warns(UserWarning) as record:
            group = prtecan.TecanfilesGroup(tecanfiles)

        def test_warn(self) -> None:
            """Warn about labelblocks anomaly."""
            assert "Different LabelblocksGroup among filenames:" in str(
                self.record[0].message
            )

        def test_labelblocksgroups(self) -> None:
            """Generate 1 LbG with .data and .metadata."""
            lbg0 = self.group.labelblocksgroups[0]
            # metadata
            assert lbg0.metadata["Number of Flashes"].value == 10.0
            assert lbg0.metadata["Gain"].value == 94
            # data
            assert lbg0.data is not None
            assert lbg0.data["A01"] == [18713.0, 17088.0, 17123.0]
            assert lbg0.data["H12"] == [28596.0, 25771.0, 28309.0]

        def test_mergeable_labelblocksgroups(self) -> None:
            """Generate 1 Lbg only with .data_normalized and only common .metadata."""
            lbg1 = self.group.labelblocksgroups[1]
            # metadata
            assert lbg1.metadata["Number of Flashes"].value == 10.0
            assert lbg1.metadata.get("Gain") is None
            assert lbg1.data is None
            # data_normalized
            np.testing.assert_almost_equal(
                lbg1.data_norm["A01"], [401.9387755, 446.9897959, 450.0]
            )
            np.testing.assert_almost_equal(
                lbg1.data_norm["H12"], [725.8163265, 693.9795918, 714.4949494]
            )

    class TestOnly1commonLbg:
        """Test TfG with different number of labelblocks, but mergeable."""

        filenames = [
            "290212_5.78.xls",  # Label1 and Label2
            "290212_20.xls",  # Label2 only
            "290212_100.xls",  # Label2 only
        ]
        tecanfiles = [prtecan.Tecanfile(data_tests / f) for f in filenames]
        # pylint: disable=W0201
        with pytest.warns(UserWarning) as record:
            group = prtecan.TecanfilesGroup(tecanfiles)

        def test_warn(self) -> None:
            """Warn about labelblocks anomaly."""
            assert "Different LabelblocksGroup among filenames" in str(
                self.record[0].message
            )

        def test_labelblocksgroups(self) -> None:
            """Generates 1 LbG with .data and .metadata."""
            lbg = self.group.labelblocksgroups[0]
            # metadata
            assert lbg.metadata["Number of Flashes"].value == 10.0
            assert lbg.metadata["Gain"].value == 93.0
            # data
            assert lbg.data is not None
            assert lbg.data["A01"] == [6289, 6462, 6465]
            assert lbg.data["H12"] == [4477, 4705, 4918]

    class TestFailToMerge:
        """Test TfG without mergeable labelblocks."""

        filenames = ["290513_5.5.xls", "290513_5.5_bad.xls"]
        tecanfiles = [prtecan.Tecanfile(data_tests / f) for f in filenames]

        def test_raise_exception(self) -> None:
            """Raise Exception when there is no way to build labelblocksGroup."""
            with pytest.raises(
                ValueError,
                match=r"No common labelblock in filenames: .*290513_5.5_bad.xls",
            ):
                prtecan.TecanfilesGroup(self.tecanfiles)


class TestTitration:
    """Test Titration class."""

    tit_ph = prtecan.Titration.fromlistfile(data_tests / "list.pH")
    tit_cl = prtecan.Titration.fromlistfile(data_tests / "list.cl20")

    @pytest.mark.filterwarnings("ignore: Different LabelblocksGroup")
    def test_conc(self) -> None:
        """It reads pH values."""
        assert self.tit_ph.conc == [5.78, 6.38, 6.83, 7.24, 7.67, 8.23, 8.82, 9.31]

    def test_labelblocksgroups(self) -> None:
        """It reads labelblocksgroups data and metadata."""
        lbg0 = self.tit_ph.labelblocksgroups[0]
        lbg1 = self.tit_ph.labelblocksgroups[1]
        # metadata
        assert lbg0.metadata["Number of Flashes"].value == 10.0
        # pH9.3 is 93 Optimal not Manual
        assert lbg1.metadata["Gain"] == prtecan.Metadata(93.0)
        # data
        assert lbg0.data is not None
        assert lbg1.data is not None
        assert lbg0.data["A01"][::2] == [30344, 31010, 33731, 37967]
        assert lbg1.data["A01"][1::2] == [9165, 15591, 20788, 22534]
        assert lbg0.data["H12"][1::2] == [20888, 21711, 23397, 25045]
        assert lbg1.data["H12"] == [4477, 5849, 7165, 8080, 8477, 8822, 9338, 9303]

    def test_labelblocksgroups_cl(self) -> None:
        """It reads labelblocksgroups data for Cl too."""
        lbg = self.tit_cl.labelblocksgroups[0]
        assert lbg.data is not None
        assert lbg.data["A01"] == [6462, 6390, 6465, 6774]
        assert lbg.data["H12"] == [4705, 4850, 4918, 5007]

    def test_export_data(self, tmp_path: Path) -> None:
        """It exports titrations data to files e.g. "A01.dat"."""
        self.tit_ph.export_data(tmp_path)
        a01 = pd.read_csv(tmp_path / "dat/A01.dat")
        h12 = pd.read_csv(tmp_path / "dat/H12.dat")
        assert a01["y1"].tolist()[1::2] == [30072, 32678, 36506, 37725]
        assert a01["y2"].tolist()[1::2] == [9165, 15591, 20788, 22534]
        assert h12["y1"].tolist()[1::2] == [20888, 21711, 23397, 25045]
        assert h12["y2"].tolist() == [4477, 5849, 7165, 8080, 8477, 8822, 9338, 9303]

    def test_raise_listfilenotfound(self) -> None:
        """It raises FileNotFoundError when list.xx file does not exist."""
        with pytest.raises(FileNotFoundError, match="Cannot find: aax"):
            prtecan.Titration.fromlistfile(Path("aax"))

    def test_bad_listfile(self) -> None:
        """It raises Exception when list.xx file is ill-shaped."""
        with pytest.raises(ValueError, match=r"Check format .* for listfile: .*"):
            prtecan.Titration.fromlistfile(data_tests / "list.pH2")


@pytest.mark.filterwarnings("ignore:OVER")
class TestTitrationAnalysis:
    """Test TitrationAnalysis class."""

    @pytest.fixture(autouse=True, scope="class")
    def titan(self) -> TitrationAnalysis:
        """Set up test class."""
        titan = prtecan.TitrationAnalysis.fromlistfile(data_tests / "140220/list.pH")
        titan.load_additions(data_tests / "140220/additions.pH")
        titan.load_scheme(data_tests / "140220/scheme.txt")
        return titan

    def test_scheme(self, titan: TitrationAnalysis) -> None:
        """It finds well position for buffer samples."""
        assert titan.scheme.buffer == ["D01", "E01", "D12", "E12"]

    def test_raise_listfilenotfound(self, titan: TitrationAnalysis) -> None:
        """It raises OSError when scheme file does not exist."""
        with pytest.raises(
            FileNotFoundError, match=r"No such file or directory: 'aax'"
        ):
            titan.load_scheme(Path("aax"))

    def test_raise_listfile_exception(self, titan: TitrationAnalysis) -> None:
        """It raises AssertionError when scheme.txt file is ill-shaped."""
        bad_schemefile = data_tests / "140220/scheme0.txt"
        msg = f"Check format [well sample] for schemefile: {bad_schemefile}"
        with pytest.raises(ValueError, match=re.escape(msg)):
            titan.load_scheme(bad_schemefile)

    def test_subtract_bg(self, titan: TitrationAnalysis) -> None:
        """It subtracts buffer average values."""
        lbg0 = titan.labelblocksgroups[0]
        lbg1 = titan.labelblocksgroups[1]
        np.testing.assert_almost_equal(
            lbg0.data_norm["E01"][::2],
            [601.72, 641.505, 674.355, 706.774],
            3,
        )
        assert lbg0.data is not None
        assert lbg0.data["E01"][::2] == [11192.0, 11932.0, 12543.0, 13146.0]
        assert lbg0.data_buffersubtracted is not None
        np.testing.assert_array_equal(
            lbg0.data_buffersubtracted["A12"][::3], [8084.5, 16621.75, 13775.0]
        )
        assert lbg1.data_buffersubtracted is not None
        np.testing.assert_array_equal(
            lbg1.data_buffersubtracted["A12"][::3], [9758.25, 1334.0, 283.5]
        )

    def test_dilution_correction(self, titan: TitrationAnalysis) -> None:
        """It applies dilution correction read from file listing additions."""
        assert titan.additions is not None
        np.testing.assert_array_equal(titan.additions, [100, 2, 2, 2, 2, 2, 2])
        assert titan.data_dilutioncorrected is not None
        assert titan.data_dilutioncorrected[1] is not None
        np.testing.assert_almost_equal(
            titan.data_dilutioncorrected[1]["A12"],
            [9758.25, 7524.795, 3079.18, 1414.04, 641.79, 402.325, 317.52],
        )

    def test_data_dilutioncorrected_norma(self, titan: TitrationAnalysis) -> None:
        """It normalizes data."""
        assert titan.data_dilutioncorrected_norm is not None
        np.testing.assert_almost_equal(
            titan.data_dilutioncorrected_norm[0]["A12"][::2],
            [434.65, 878.73, 975.58, 829.46],
            2,
        )
        np.testing.assert_almost_equal(
            titan.data_dilutioncorrected_norm[1]["A12"][::2],
            [871.272, 274.927, 57.303, 28.35],
            3,
        )

    def test_keys(self, titan: TitrationAnalysis) -> None:
        """It gets well positions for ctrl and unknown samples."""
        assert set(titan.scheme.names) == {"NTT", "G03", "V224Q", "S202N"}
        x = {"B12", "H12", "F01", "C12", "F12", "C01", "H01", "G12", "B01", "G01"}
        assert set(titan.scheme.ctrl) - {"A01", "A12"} == x

    @pytest.mark.skipif(sys.platform == "win32", reason="broken on windows")
    def test_fit(self, titan: TitrationAnalysis) -> None:
        """It fits each label separately."""
        titan.fit("pH", nrm=True, bg=True, dil=True)
        fit0 = titan.fittings[0].sort_index()
        fit1 = titan.fittings[1].sort_index()
        df0 = pd.read_csv(data_tests / "140220/fit0.csv", index_col=0)
        df1 = pd.read_csv(data_tests / "140220/fit1.csv", index_col=0)
        pd.testing.assert_frame_equal(df0.sort_index(), fit0, rtol=1e0)
        pd.testing.assert_frame_equal(
            df1,
            fit1,
            check_like=True,
            check_categorical=False,
            atol=1e-3,
        )
        # fit up to the second-last data point
        titan.fit("pH", fin=-1, nrm=True, bg=True, dil=True)
        fit0 = titan.fittings[0].sort_index()
        fit1 = titan.fittings[1].sort_index()
        df0 = pd.read_csv(data_tests / "140220/fit0-1.csv", index_col=0)
        df1 = pd.read_csv(data_tests / "140220/fit1-1.csv", index_col=0)
        pd.testing.assert_frame_equal(
            df0.sort_index()[1:], fit0[1:], check_like=True, atol=1e-4
        )
        pd.testing.assert_frame_equal(df1.sort_index(), fit1, atol=1e-5)



#!/usr/bin/env python

import os
import argparse
import numpy as np
import pandas as pd
from collections import namedtuple
from scipy import optimize
import matplotlib.pyplot as plt

import seaborn


def main():
    """Fit pH and cl titrations where data are spectra.
    input: spectra_table.csv; _note
    output: pK; spK; and plot (*.pdf)
    """
    description = "Fit spectra from a pH or a Cl titration"
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('csvtable',
                        help="a table containing spectra arranged in columns")
    parser.add_argument('note_file',
                        help="a file to describe the titration")
    parser.add_argument("-d", '--out', dest='out',
                        help="destination directory (default: .)")
    parser.add_argument("-m", "--method-of-analysis", action="store",
                        default="svd", choices=["svd", "band"],
                        dest='analysis_method')
    parser.add_argument("-t", "--titration-of", action="store", default="pH",
                        choices=["pH", "cl"], dest='titration_type')
    parser.add_argument("-b", "--band-interval", action="store", nargs=2,
                        type=int, dest='band',
                        help='Integration interval from <1> to <2>')
    parser.add_argument("-v", "--verbose", action="store_true",
                        help="increase output verbosity")
    args = parser.parse_args()
    # input of spectra (csv) and titration (note) data
    csv = pd.read_csv(args.csvtable)
    note_file = pd.read_table(args.note_file)
    note_file = note_file[note_file['mutant'] != 'buffer']
    # TODO aggregation logic for some pH or cloride
    Note = namedtuple("note", "wells conc")
    # pH or Cl titration
    if args.titration_type == "cl":
        note = Note(list(note_file.well), list(note_file.Cl))

        def fz(Kd, p, x):
            return ((p[0] + p[1] * x / Kd) / (1 + x / Kd))
    elif args.titration_type == "pH":
        note = Note(list(note_file.well), list(note_file.pH))

        def fz(pK, p, x):
            return (p[1] + p[0] * 10 ** (pK - x)) / (1 + 10 ** (pK - x))
    df = csv[note.wells]
    df.index = csv['lambda']
    conc = np.array(note.conc)
    # sideeffect print input data
    if args.verbose:
        print(csv)
        print(note)
        print('conc vector\n', conc)
        print('DataFrame\n', df)
    if args.analysis_method == 'svd':
        # svd on difference spectra
        ddf = df.sub(df.iloc[:, 0], axis=0)
        u, s, v = np.linalg.svd(ddf)
        # fitting
        result = fit_titration(fz, conc, v[0, :])
        # plotting
        seaborn.set_style('ticks')
        fig1 = plt.figure(figsize=(12, 8))
        ax1 = fig1.add_axes([0.05, 0.65, 0.32, 0.31])
        plt.grid(True)
        ax2 = fig1.add_axes([0.42, 0.65, 0.32, 0.31])
        plt.grid(True)
        ax1.plot(df)
        ax2.plot(ddf.index, u[:, 0], 'k-', lw=3)
        ax2.plot(ddf.index, u[:, 1], 'b--')
        ax3 = fig1.add_axes([0.80, 0.65, 0.18, 0.31], yscale='log',
                            xticks=[1, 2, 3, 4], title='autovalues')
        ax3.bar([1, 2, 3, 4], (s**2 / sum(s**2))[:4],
                align='center', alpha=0.7, width=0.66)
        ax4 = fig1.add_axes([0.05, 0.08, 0.50, 0.50], title="fitting")
        ax5 = fig1.add_axes([0.63, 0.08, 0.35, 0.50],
                            title='SVD coefficients',
                            xlabel='1$^{st}$ autovector',
                            ylabel='2$^{nd}$ autovector')
        ax4.scatter(conc, v[0, :])
        xmin = conc.min()
        xmax = conc.max()
        xmax += (xmax - xmin) / 7
        xlin = np.linspace(xmin, xmax, 100)
        ax4.plot(xlin, fz(result.K, [result.SA, result.SB], xlin))
        result.s1 = str(round(result.K, 2)) + ' \u00B1 '
        result.s1 = result.s1 + str(round(result.sK, 2))
        plt.figtext(.26, .54, result.s1, size=20)
        ax5.plot(v[:, 1], v[:, 2], lw=0.8)
        for x, y, l in zip(v[:, 1], v[:, 2], note.wells):
            ax5.text(x, y, l)
        if 1 == 1:
            kd = []
            sa = []
            sb = []
            for i in range(100):
                boot_idxs = np.random.randint(0, len(ddf.columns) - 1,
                                                      len(ddf.columns))
                ddf2 = df.iloc[:, boot_idxs]
                conc2 = conc[boot_idxs]
                u, s, v = np.linalg.svd(ddf2)
                result2 = fit_titration(fz, conc2, v[0, :])
                kd.append(result2.K)
                sa.append(result2.SA)
                sb.append(result2.SB)
            bs = pd.DataFrame({'kd': kd, 'SA': sa, 'SB': sb})
            bs.to_csv('bs.txt')
    elif args.analysis_method == 'band':
        # fitting
        try:
            ini = args.band[0]
            fin = args.band[1]
            y = []
            for c in df.columns:
                y.append(df[c].loc[ini: fin].sum())
            result = fit_titration(fz, conc, y)
        except:
            print('''bands [{0}, {1}] not in index.
                  Try other values'''.format(ini, fin))
            raise
        # plotting
        fig1 = plt.figure(figsize=(12, 8))
        ax4 = fig1.add_axes([0.05, 0.08, 0.50, 0.50], title="fitting")
        ax4.scatter(conc, y)
        xmin = conc.min()
        xmax = conc.max()
        xmax += (xmax - xmin) / 7
        xlin = np.linspace(xmin, xmax, 100)
        ax4.plot(xlin, fz(result.K, [result.SA, result.SB], xlin))
        result.s1 = str(round(result.K, 2)) + ' \u00B1 '
        result.s1 = result.s1 + str(round(result.sK, 2)) + '[' + str(ini)
        result.s1 = result.s1 + ':' + str(fin) + ']'
        plt.figtext(.26, .54, result.s1, size=20)
        for x, y, l in zip(conc, y, note.wells):
            ax4.text(x, y, l)

    # output
    f_csv_shortname = os.path.splitext(os.path.split(args.csvtable)[1])[0]
    f_note_shortname = os.path.splitext(os.path.split(args.note_file)[1])[0]
    f_out = f_csv_shortname.join([args.analysis_method + "_", "_"])
    f_out = f_out + f_note_shortname
    if args.out:
        if not os.path.isdir(args.out):
            os.makedirs(args.out)
        f_out = os.path.join(args.out, f_out)
    fig1.savefig(f_out + ".pdf")

    print("best-fitting using: ", args.analysis_method)
    print("spectra csv file: ", f_csv_shortname)
    print("note file: ", f_note_shortname)
    print("K = ", round(result.K, 3))
    print("sK = ", round(result.sK, 3))
    print("SA = ", round(result.SA, 3))
    print("sSA = ", round(result.sSA, 3))
    print("SB = ", round(result.SB, 3))
    print("sSB = ", round(result.sSB, 3))


def fit_titration(fz, x, y):
    """Fit a dataset (x, y) using a single-site binding model provided by the
    function **fz** that defines a constant *K* and 2 plateau *SA* and *SB*.

    Parameters
    ----------

    x, y : numpy arrays

    Return
    ------

    res : namedtuple
        contains the least square results

    """
    def ssq(p, x, y1):
        return np.r_[y1 - fz(p[0], p[1:3], x)]
    # plateau calculation
    df = pd.DataFrame({'x': x, 'y': y})
    SA = df.y[df.x == min(df.x)].values[0]
    SB = df.y[df.x == max(df.x)].values[0]
    p0 = np.r_[7.1, SA, SB]
    p, cov, info, msg, success = optimize.leastsq(ssq, p0, args=(x, y),
                                                  full_output=True, xtol=1e-11)
    # ftol=1.49012e-23,
    res = namedtuple("Result", "success msg df chisqr K sK SA sSA SB sSB")
    res.msg = msg
    res.success = success
    if 1 <= success <= 4:
        chisq = sum(info['fvec'] * info['fvec'])
        res.df = len(y) - len(p)
        res.chisqr = chisq / res.df
        res.K = p[0]
        res.sK = np.sqrt(cov[0][0] * res.chisqr)
        res.SA = p[1]
        res.sSA = np.sqrt(cov[1][1] * res.chisqr)
        res.SB = p[2]
        res.sSB = np.sqrt(cov[2][2] * res.chisqr)
    return res


main()



"""
Parse Tecan files and fit files that are grouped into titrations.

A 'titration' in this context is defined within a list.pH or list.cl file. These files describe the properties and parameters of the titration experiment.

The command-line interface 'prtecan' is used to construct 96 distinct titrations and export the results to .txt files. This allows for further analysis or visualization of the titration data.

In cases where two label blocks exist in the Tecan files, the module performs a 'global fitting' operation. Global fitting is a method used in data analysis to best fit a model to multiple data sets simultaneously, under the assumption that some underlying parameters are shared between the data sets.

Please consult the documentation for each function in this module for more detailed information on their operation and use.

"""
from clophfit.prtecan.prtecan import FitFirstError
from clophfit.prtecan.prtecan import Labelblock
from clophfit.prtecan.prtecan import LabelblocksGroup
from clophfit.prtecan.prtecan import Metadata
from clophfit.prtecan.prtecan import PlateScheme
from clophfit.prtecan.prtecan import Tecanfile
from clophfit.prtecan.prtecan import TecanfilesGroup
from clophfit.prtecan.prtecan import Titration
from clophfit.prtecan.prtecan import TitrationAnalysis
from clophfit.prtecan.prtecan import calculate_conc
from clophfit.prtecan.prtecan import dilution_correction
from clophfit.prtecan.prtecan import extract_metadata
from clophfit.prtecan.prtecan import lookup_listoflines
from clophfit.prtecan.prtecan import merge_md
from clophfit.prtecan.prtecan import read_xls
from clophfit.prtecan.prtecan import strip_lines

__all__ = [
    "FitFirstError",
    "Labelblock",
    "LabelblocksGroup",
    "Metadata",
    "PlateScheme",
    "Tecanfile",
    "TecanfilesGroup",
    "Titration",
    "TitrationAnalysis",
    "calculate_conc",
    "dilution_correction",
    "extract_metadata",
    "lookup_listoflines",
    "merge_md",
    "read_xls",
    "strip_lines",
]



#!/usr/bin/python
#

import pandas as pd
import numpy as np
import sys

try:
    list_file = str(sys.argv[1])
    b = pd.read_table(list_file, header=None)
    b.columns = ['file']
    files = []
    for f in b['file']:
        tmp = pd.read_table(f, header=None, skiprows=2)
        tmp.columns = ['lambda', 'F']
        files.append(tmp)
    df = pd.DataFrame(
            np.transpose([f.F for f in files]),
            columns=['s'+str(i) for i in range(len(files))],
            index=files[0]['lambda'])
    df.to_csv('table.csv')

except:
    print('merge need a file named "list"')
    print(sys.argv[1])
    print(list_file)



#!/usr/bin/python
#

import argparse
import os
import pandas as pd


corr = {'pH': [1, 1.02, 1.04, 1.06, 1.08, 1.1, 1.12],
        'cl': [1.000000, 1.017857, 1.035714, 1.053571, 1.071429, 1.089286,
               1.107143, 1.160714, 1.196429]}

corr = {'pH': [1, 1.02, 1.04, 1.06, 1.08, 1.1],
        'cl': [1.000000, 1.017857, 1.035714, 1.053571, 1.071429, 1.089286,
               1.107143, 1.160714, 1.196429]}


def main():
    description = "make correction for dilution"
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('ffile',
                        help="a file x y1 y2 --> y1*corr y2*corr")
    parser.add_argument('out',
                        help="destination directory")
    parser.add_argument('buffer', nargs=2,
                        help="buffer values for y1 y2")
    parser.add_argument("-t", "--titration-of", action="store", default="pH",
                        choices=["pH", "cl"], dest='titration_type')
    args = parser.parse_args()
    df = pd.read_csv(args.ffile)
    df.y1 -= float(args.buffer[0])
    df.y2 -= float(args.buffer[1])
    df.y1 *= corr[args.titration_type]
    df.y2 *= corr[args.titration_type]
    if not os.path.isdir(args.out):
        os.makedirs(args.out)
    df.to_csv(os.path.join(args.out, args.ffile), index=False)


main()



"""ClopHfit: Parse plate-reader and fit ClopHensor titrations."""
from pathlib import Path

from pkg_resources import get_distribution  # type: ignore

__version__ = get_distribution("clophfit").version
__default_enspire_out_dir__ = Path(f"Meas-{__version__}")



"""Test ``clop`` cli."""
from __future__ import annotations

import filecmp
from pathlib import Path

import pytest
from click.testing import CliRunner
from matplotlib.testing.compare import compare_images  # type: ignore
from matplotlib.testing.exceptions import ImageComparisonFailure  # type: ignore

from clophfit.__main__ import clop

# tests path
tpath = Path(__file__).parent


def test_eq1() -> None:
    """It runs XXX pr.tecan and generates correct results and graphs."""
    runner = CliRunner()
    result = runner.invoke(clop, ["eq1", "2", "2", "2"])
    assert result.exit_code == 0
    assert "4." in result.output


def test_prenspire(tmp_path: Path) -> None:
    """Test prenspire command with actual data and validate output."""
    expected = tpath / "EnSpire" / "cli" / "output"
    input_csv = tpath / "EnSpire" / "cli" / "NTT_37C_pKa.csv"
    out = tmp_path / "E"
    out.mkdir()
    runner = CliRunner()
    result = runner.invoke(
        clop,
        ["pr.enspire", str(input_csv), "--out", str(out)],
    )
    assert result.exit_code == 0
    # validate output files
    assert (out / "NTT_37C_pKa_A.csv").exists()
    assert (out / "NTT_37C_pKa_B.csv").exists()
    assert (out / "NTT_37C_pKa_A.png").exists()
    assert (out / "NTT_37C_pKa_B.png").exists()
    # validate output file contents
    assert filecmp.cmp(out / "NTT_37C_pKa_A.csv", expected / "NTT_37C_pKa_A.csv")
    assert filecmp.cmp(out / "NTT_37C_pKa_B.csv", expected / "NTT_37C_pKa_B.csv")
    # validate graph
    for f in ["NTT_37C_pKa_A.png", "NTT_37C_pKa_B.png"]:
        msg = compare_images(out / f, expected / f, 0.0001)
        if msg:  # pragma: no cover
            raise ImageComparisonFailure(msg)


@pytest.mark.filterwarnings("ignore:OVER")
def test_prtecan(tmp_path: Path) -> None:
    """Test prtecan command with actual data."""
    lfile = str(tpath / "Tecan" / "140220" / "list.pH")
    sfile = str(tpath / "Tecan" / "140220" / "scheme.txt")
    out = tmp_path / "out3"
    out.mkdir()
    runner = CliRunner()
    result = runner.invoke(
        clop,
        ["pr.tecan", lfile, "--out", str(out), "--fit", "--scheme", sfile, "--bg"],
    )
    assert result.exit_code == 0


@pytest.mark.parametrize(
    ("csv_file", "expected_output", "additional_params"),
    [
        ("A04 Cl_A.csv", " K :  13.64  14.28  14.86  15.44  16.04  16.71  17.51", None),
        ("A04 Cl_A.csv", " K :  13.52  14.11  14.66  15.20", ["-b", "480", "530"]),
        ("A04 Cl_B.csv", " K :  12.73  13.46  14.14  14.82  15.54  16.35  17.32", None),
        ("A04 Cl_B.csv", " K :  12.93  13.66  14.34  15.02", ["-b", "480", "530"]),
    ],
)
def test_fit_titration(
    csv_file: str, expected_output: str, additional_params: list[str], tmp_path: Path
) -> None:
    """Test the old ``fit_titration.py`` script with actual data and validate output."""
    note_fp = tpath / "data" / "NTT-A04-Cl_note"
    csv_fp = tpath / "data" / "Meas" / csv_file
    out = tmp_path / "tit_fit"
    out.mkdir()
    runner = CliRunner()
    base_args = ["fit_titration", str(csv_fp), str(note_fp)]
    base_args.extend(["-t", "cl", "-d", str(out)])
    if additional_params:
        base_args.extend(additional_params)
        sbands = f"({additional_params[1]}, {additional_params[2]})"
    else:
        sbands = "None"
    result = runner.invoke(clop, base_args)
    print(result.exc_info)
    print(result.output)
    assert result.exit_code == 0
    assert expected_output in result.output
    # Asserting that PDF is created
    expected_pdf_filename = out / f"{csv_fp.stem}_{sbands}_{note_fp.stem}.pdf"
    assert expected_pdf_filename.exists()



"""Fit Cl binding and pH titration."""
from __future__ import annotations

import typing
from typing import Sequence

import lmfit  # type: ignore
import matplotlib as mpl  # type: ignore
import matplotlib.colors as mcolors  # type: ignore
import matplotlib.pyplot as plt  # type: ignore
import numpy as np
import pandas as pd
import scipy  # type: ignore
import scipy.stats  # type: ignore
import seaborn as sb  # type: ignore # noqa: ICN001
from numpy.typing import NDArray
from uncertainties import ufloat  # type: ignore

COLOR_MAP = plt.cm.Set1


@typing.overload
def _binding_pk(x: float, K: float, S0: float, S1: float) -> float:  # noqa: N803
    ...


@typing.overload
def _binding_pk(
    x: NDArray[np.float_], K: float, S0: float, S1: float  # noqa: N803
) -> NDArray[np.float_]:
    ...


def _binding_pk(
    x: float | NDArray[np.float_], K: float, S0: float, S1: float  # noqa: N803
) -> float | NDArray[np.float_]:
    return S0 + (S1 - S0) * 10 ** (K - x) / (1 + 10 ** (K - x))


@typing.overload
def _binding_kd(x: float, K: float, S0: float, S1: float) -> float:  # noqa: N803
    ...


@typing.overload
def _binding_kd(
    x: NDArray[np.float_], K: float, S0: float, S1: float  # noqa: N803
) -> NDArray[np.float_]:
    ...


def _binding_kd(
    x: float | NDArray[np.float_], K: float, S0: float, S1: float  # noqa: N803
) -> float | NDArray[np.float_]:
    return S1 + (S0 - S1) * x / K / (1 + x / K)


def kd(
    kd1: float, pka: float, ph: NDArray[np.float_] | float
) -> NDArray[np.float_] | float:
    """Infinite cooperativity model.

    It can describe pH-dependence for chloride dissociation constant.

    Parameters
    ----------
    kd1 : float
        Dissociation constant at pH <= 5.0 (fully protonated).
    pka : float
        Acid dissociation constant.
    ph : Xtype
        pH value(s).

    Returns
    -------
    Xtype
        Predicted Kd value(s).

    Examples
    --------
    >>> kd(10, 8.4, 7.4)
    11.0
    >>> import numpy as np
    >>> kd(10, 8.4, np.array([7.4, 8.4]))
    array([11., 20.])

    """
    return kd1 * (1 + 10 ** (pka - ph)) / 10 ** (pka - ph)


# TODO other from datan
# TODO: use this like fz in prtecan
def fz_kd_singlesite(
    k: float, p: NDArray[np.float_] | Sequence[float], x: NDArray[np.float_]
) -> NDArray[np.float_]:
    """Fit function for Cl titration."""
    return (float(p[0]) + float(p[1]) * x / k) / (1 + x / k)


def fz_pk_singlesite(
    k: float, p: NDArray[np.float_] | Sequence[float], x: NDArray[np.float_]
) -> NDArray[np.float_]:
    """Fit function for pH titration."""
    return (float(p[1]) + float(p[0]) * 10 ** (k - x)) / (1 + 10 ** (k - x))


# `fit_titration` is exported and use in prtecan.
def fit_titration(  # noqa: PLR0913, PLR0915
    kind: str,
    x: Sequence[float],
    y: NDArray[np.float_],
    y2: NDArray[np.float_] | None = None,
    residue: NDArray[np.float_] | None = None,
    residue2: NDArray[np.float_] | None = None,
    tval_conf: float = 0.95,
) -> pd.DataFrame:
    """Fit pH or Cl titration using a single-site binding model.

    Returns confidence interval (default=0.95) for fitting params (cov*tval), rather than
    standard error of the fit. Use scipy leastsq. Determine 3 fitting parameters:
    - binding constant *K*
    - and 2 plateau *SA* and *SB*.

    Parameters
    ----------
    kind : str
        Titration type {'pH'|'Cl'}
    x : Sequence[float]
        Dataset x-values.
    y : NDArray[np.float]
        Dataset y-values.
    y2 : NDArray[np.float], optional
        Optional second dataset y-values (share x with main dataset).
    residue : NDArray[np.float], optional
        Residues for main dataset.
    residue2 : NDArray[np.float], optional
        Residues for second dataset.
    tval_conf : float
        Confidence level (default 0.95) for parameter estimations.

    Returns
    -------
    pd.DataFrame
        Fitting results.

    Raises
    ------
    NameError
        When kind is different than "pH" or "Cl".

    Examples
    --------
    >>> import numpy as np
    >>> fit_titration("Cl", np.array([1.0, 10, 30, 100, 200]), \
          np.array([10, 8, 5, 1, 0.1]))[["K", "sK"]]
               K         sK
    0  38.955406  30.201929

    """
    if kind == "pH":
        fz = fz_pk_singlesite
    elif kind == "Cl":
        fz = fz_kd_singlesite
    else:
        msg = "kind= pH or Cl"
        raise NameError(msg)

    def compute_p0(x: Sequence[float], y: NDArray[np.float_]) -> NDArray[np.float_]:
        data = pd.DataFrame({"x": x, "y": y})
        p0sa = data.y[data.x == min(data.x)].to_numpy()[0]
        p0sb = data.y[data.x == max(data.x)].to_numpy()[0]
        p0k = np.average([max(y), min(y)])
        try:
            x1, y1 = data[data["y"] >= p0k].to_numpy()[0]
        except IndexError:
            x1 = np.nan
            y1 = np.nan
        try:
            x2, y2 = data[data["y"] <= p0k].to_numpy()[0]
        except IndexError:
            x2 = np.nan
            y2 = np.nan
        p0k = (x2 - x1) / (y2 - y1) * (p0k - y1) + x1
        return np.array(np.r_[p0k, p0sa, p0sb])

    if y2 is None:

        def ssq1(
            p: NDArray[np.float_], x: NDArray[np.float_], y1: NDArray[np.float_]
        ) -> NDArray[np.float_]:
            return np.array(np.r_[y1 - fz(p[0], p[1:3], x)])

        p0 = compute_p0(x, y)
        p, cov, info, msg, success = scipy.optimize.leastsq(
            ssq1, p0, args=(np.array(x), y), full_output=True, xtol=1e-11
        )
    else:

        def ssq2(  # noqa: PLR0913
            p: NDArray[np.float_],
            x: NDArray[np.float_],
            y1: NDArray[np.float_],
            y2: NDArray[np.float_],
            rd1: NDArray[np.float_],
            rd2: NDArray[np.float_],
        ) -> NDArray[np.float_]:
            return np.array(
                np.r_[
                    (y1 - fz(p[0], p[1:3], x)) / rd1**2,
                    (y2 - fz(p[0], p[3:5], x)) / rd2**2,
                ]
            )

        p1 = compute_p0(x, y)
        p2 = compute_p0(x, y2)
        ave = np.average([p1[0], p2[0]])
        p0 = np.r_[ave, p1[1], p1[2], p2[1], p2[2]]
        tmp = scipy.optimize.leastsq(
            ssq2,
            p0,
            full_output=True,
            xtol=1e-11,
            args=(np.array(x), y, y2, residue, residue2),
        )
        p, cov, info, msg, success = tmp
    res = pd.DataFrame({"ss": [success]})
    res["msg"] = msg
    if 1 <= success <= 4:  # noqa: PLR2004
        try:
            tval = (tval_conf + 1) / 2
            chisq = sum(info["fvec"] * info["fvec"])
            res["df"] = len(y) - len(p)
            res["tval"] = scipy.stats.distributions.t.ppf(tval, res.df)
            res["chisqr"] = chisq / res.df
            res["K"] = p[0]
            res["SA"] = p[1]
            res["SB"] = p[2]
            if y2 is not None:
                res["df"] += len(y2)
                res["tval"] = scipy.stats.distributions.t.ppf(tval, res.df)
                res["chisqr"] = chisq / res.df
                res["SA2"] = p[3]
                res["SB2"] = p[4]
                res["sSA2"] = np.sqrt(cov[3][3] * res.chisqr) * res.tval
                res["sSB2"] = np.sqrt(cov[4][4] * res.chisqr) * res.tval
            res["sK"] = np.sqrt(cov[0][0] * res.chisqr) * res.tval
            res["sSA"] = np.sqrt(cov[1][1] * res.chisqr) * res.tval
            res["sSB"] = np.sqrt(cov[2][2] * res.chisqr) * res.tval
        except TypeError:
            pass  # if some params are not successfully determined.
    return res


###############
def fit_binding(
    x: NDArray[np.float_],
    y: NDArray[np.float_],
    model_func: typing.Callable[[float, float, float, float], float],
) -> lmfit.model.ModelResult:
    """Fit a dataset (x, y) using a single-site binding model with lmfit.

    The model is provided by the `model_func` argument, a function defining the
    model for fitting, which takes four arguments - x, K, S0 and S1, and returns
    a float.

    Parameters
    ----------
    x : NDArray[np.float]
        The x values of the dataset.
    y : NDArray[np.float]
        The y values of the dataset.
    model_func : Callable
        The function defining the model for fitting.

    Returns
    -------
    result : ModelResult
        lmfit's object containing the fitting results.

    Raises
    ------
    ValueError
        If the optimization fails.
    """
    model = lmfit.Model(model_func)
    params = lmfit.Parameters()
    # Data-driven initialization
    xydata = pd.DataFrame({"x": x, "y": y})
    s0_initial = y[0]
    s1_initial = y[-1]
    target_y = (s1_initial + s0_initial) / 2
    k_initial = xydata.loc[(xydata.y - target_y).abs().idxmin(), "x"]
    # Parameters initialization
    params.add("K", value=k_initial, min=2)
    params.add("S1", value=s1_initial)
    params.add("S0", value=s0_initial)
    result = model.fit(y, params, x=x)
    if result.success is not True:
        message = f"Optimization failed with message: {result.message}"
        raise ValueError(message)
    return result


def plot_lmfit(
    ax: mpl.axes.Axes,
    result: lmfit.model.ModelResult,
    hue_norm: tuple[float, float],
    palette: str,
) -> None:
    """Plot the results of the lmfit model on a given axes.

    This function creates a scatter plot of the data used for fitting, then
    overlays the best-fit curve and its uncertainty band. The data points are
    colored according to their x-value, and a text element is added to the plot
    to display the fitted parameter K with its uncertainty.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes on which to plot the lmfit results.
    result : lmfit.model.ModelResult
        The lmfit ModelResult object containing the fit results.
    hue_norm : tuple[float, float]
        A tuple of two floats specifying the normalization range for the colormap.
    palette : str
        The name of the colormap to use for coloring the data points.
    """
    x = result.userkws["x"]
    y = result.data
    n0 = hue_norm[0]
    n1 = hue_norm[1]
    ax.scatter(x, y, c=x, cmap=palette, vmin=n0, vmax=n1, s=99, edgecolors="k")
    # Adjust x-axis limits
    xmin = x.min()
    xmax = x.max()
    xmax += (xmax - xmin) / 7
    xfit = np.linspace(xmin, xmax, 100)
    # Plot fit line
    best_fit = result.eval(x=xfit)
    ax.plot(xfit, best_fit, "k--")
    dely = result.eval_uncertainty(x=xfit)
    ax.fill_between(xfit, best_fit - dely, best_fit + dely, color="#FEDCBA", alpha=0.5)
    # Add fit result to plot
    k = ufloat(result.params["K"].value, result.params["K"].stderr)
    title = "=".join(["K", str(k).replace("+/-", "±")])
    sb.mpl.pyplot.figtext(0.26, 0.54, title, size=20)


def plot_spectra(
    f: pd.DataFrame,
    ax: mpl.axes.Axes,
    hue_norm: tuple[float, float],
    palette: str,
    kind: str,
) -> None:
    """Plot spectra.

    Parameters
    ----------
    f : pd.DataFrame
        The DataFrame containing spectral data.
    ax : mpl.axes.Axes
        The Axes object to which the plot should be added.
    hue_norm : Tuple[float, float]
        The normalization for the hue mapping, as a tuple of the form (min, max).
    palette : str
        The name of the palette to use for the hue mapping.
    kind : str
        The variable in `data` to map plot aspects to.
    """
    color_map = plt.get_cmap(palette)
    normalize = mcolors.Normalize(vmin=hue_norm[0], vmax=hue_norm[1])
    for i in range(len(f.columns)):
        ax.plot(f.index, f.iloc[:, i], color=color_map(normalize(f.columns[i])))
    _apply_common_plot_style(ax, "Spectra", "Wavelength", "Fluorescence")
    # Add a colorbar for reference
    sm = plt.cm.ScalarMappable(cmap=color_map, norm=normalize)
    sm.set_array([])
    plt.colorbar(sm, ax=ax, label=kind)


def _apply_common_plot_style(
    ax: mpl.axes.Axes,
    title: str,
    xlabel: str,
    ylabel: str,
) -> None:
    """Apply grid style and add title and labels."""
    ax.grid(True)
    ax.set_title(title)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)


def _plot_autovectors(wl: pd.Index, u: NDArray[np.float_], ax: mpl.axes.Axes) -> None:
    """Plot autovectors.

    Parameters
    ----------
    wl : pd.Index
        The index of spectra data frame.
    u : np.ndarray
        The left singular vectors obtained from SVD.
    ax : mpl.axes.Axes
        The Axes object to which the plot should be added.

    """
    number_autovectors = 4
    for i in range(number_autovectors):
        ax.plot(wl, u[:, i], color=COLOR_MAP(i), lw=3 / (i + 1), alpha=(1 - 0.2 * i))
    _apply_common_plot_style(ax, "Autovectors", "Wavelength", "Magnitude")


def plot_autovalues(s: NDArray[np.float_], ax: mpl.axes.Axes) -> None:
    """Plot the singular values from SVD.

    Parameters
    ----------
    s : numpy.ndarray
        The singular values from the SVD.
    ax : mpl.axes.Axes
        The axes on which to plot the singular values.
    """
    data = pd.DataFrame({"index": range(1, len(s) + 1), "singular_values": s})
    sb.scatterplot(
        x="index",
        y="singular_values",
        data=data,
        ax=ax,
        hue="index",
        s=99,
        legend=False,
        palette=COLOR_MAP.name,
    )
    _apply_common_plot_style(ax, "Singular Values from SVD", "Index", "Singular Value")
    ax.set(yscale="log")
    ax.set_xticks(np.arange(1, len(s) + 1))


def plot_pca(
    v: NDArray[np.float_],
    ax: mpl.axes.Axes,
    conc: NDArray[np.float_],
    hue_norm: tuple[float, float],
    palette: str,
) -> None:
    """Plot the first two principal components.

    Parameters
    ----------
    v : NDArray[np.float_]
        The matrix containing the principal components.
    ax : mpl.axes.Axes
        The Axes object to which the plot should be added.
    conc : NDArray[np.float_]
        The concentrations used for the titration.
    hue_norm : Tuple[float, float]
        The normalization for the hue mapping.
    palette : str
        The colormap to be used.

    """
    n0 = hue_norm[0]
    n1 = hue_norm[1]
    ax.scatter(v[1], v[0], c=conc, cmap=palette, vmin=n0, vmax=n1, s=99, edgecolors="k")
    _apply_common_plot_style(ax, "PCA plot", "", "")
    ax.set_ylabel("First Principal Component", color=COLOR_MAP(0))
    ax.set_xlabel("Second Principal Component", color=COLOR_MAP(1))
    # Add labels.
    for x, y, w in zip(v[1], v[0], conc):
        ax.text(x, y, w)


def analyze_spectra_glob(
    titration: dict[str, pd.DataFrame],
    kind: str,
    dbands: dict[str, tuple[int, int]] | None = None,
    x_combined: dict[str, NDArray[np.float_]] | None = None,
    y_combined: dict[str, NDArray[np.float_]] | None = None,
) -> (
    tuple[
        plt.Figure | None,
        lmfit.model.ModelResult | None,
        plt.Figure,
        lmfit.model.MinimizerResult,
    ]
    | tuple[
        plt.Figure,
        lmfit.model.ModelResult,
        plt.Figure | None,
        lmfit.model.MinimizerResult | None,
    ]
):
    """Analyze multi-label spectra visualize the results."""
    _gap_ = 1
    dbands = dbands or {}
    x_combined = x_combined or {}
    y_combined = y_combined or {}
    labels_svd = titration.keys() - dbands.keys()
    if len(labels_svd) > 1:
        # Concatenate spectra.
        prev_max = 0
        adjusted_list = []
        for label in labels_svd:
            spectra_adjusted = titration[label].copy()  # Avoid modifying original data
            spectra_adjusted.index += prev_max - spectra_adjusted.index.min() + _gap_
            prev_max = spectra_adjusted.index.max()
            adjusted_list.append(spectra_adjusted)
        spectra_merged = pd.concat(adjusted_list)
        # Analyze concatenated spectra.
        figure_svd, result_svd = analyze_spectra(spectra_merged, kind)
    else:
        figure_svd, result_svd = (None, None)
    if len(dbands.keys()) > 1:
        params = lmfit.Parameters()
        params.add("K", value=7, min=0)
        for label in dbands:
            params.add(f"S0_{label}", value=y_combined[label][0])
            params.add(f"S1_{label}", value=y_combined[label][-1])
        xc = {label: x_combined[label] for label in dbands}
        yc = {label: y_combined[label] for label in dbands}
        fz = _binding_kd if kind == "Cl" else _binding_pk
        result_bands = lmfit.minimize(_binding_residuals, params, args=(fz, xc, yc))
        figure_bands, ax = plt.subplots()
        xfit = {label: np.linspace(x.min(), x.max(), 100) for label, x in xc.items()}
        yfit = _binding_residuals(result_bands.params, fz, xfit)
        for label in yc:
            ax.plot(xc[label], yc[label], "o", xfit[label], yfit[label], "-")
        ax.grid(True)
    else:
        figure_bands, result_bands = (None, None)
    return figure_svd, result_svd, figure_bands, result_bands


def analyze_spectra(
    spectra: pd.DataFrame, kind: str, band: tuple[int, int] | None = None
) -> tuple[plt.Figure, lmfit.model.ModelResult]:
    """Analyze spectral data and visualize the results.

    Depending on the method chosen, this function either performs SVD on the input
    data, generating principal component spectra and plotting the resulting data,
    or integrates the spectral data over a specified band and fits the integrated
    data to a binding model.

    The 'kind' parameter determines the binding model used for fitting as well as
    the colormap and hue normalization for the plots.

    Parameters
    ----------
    spectra : pd.DataFrame
        A DataFrame where each column represents a spectrum at a different condition
        (for instance, a different pH or ligand concentration) and each row
        corresponds to a different wavelength.

    kind : str
        Specifies the type of data either 'pH' or 'Cl' titrations.

    band : tuple[int, int], optional
        The band to integrate over for the band method. If None (default), performs SVD.
        The tuple values correspond to the start and end wavelengths for the band.


    Returns
    -------
    fig : matplotlib.figure.Figure
        A Figure object with the plots visualizing the analysis results.

    result : lmfit.model.ModelResult
        The result of the lmfit model fitting.

    Raises
    ------
    ValueError
        If the band parameters are not in the spectra's index when the band method
        is used.

    Notes
    -----
    For the SVD method, the function plots the original spectra, the principal
    component vectors, the singular values, the first principal component, and
    the fitting results. For the band method, it plots the original spectra and
    the fitting results.
    """
    if kind == "Cl":
        hue_norm = (0.0, 200.0)
        palette = sb.cm.crest.name
        fz = _binding_kd
    else:
        hue_norm = (5.7, 8.7)
        palette = sb.cm.vlag_r.name
        fz = _binding_pk
    x = spectra.columns.to_numpy()
    fig = sb.mpl.pyplot.figure(figsize=(12, 8))
    ax1 = fig.add_axes([0.05, 0.65, 0.32, 0.31])
    plot_spectra(spectra, ax1, hue_norm, palette, kind)
    if band is None:  # SVD
        ddf = spectra.sub(spectra.iloc[:, 0], axis=0)
        u, s, v = np.linalg.svd(ddf)
        y = v[0, :] + 1
        result = fit_binding(x, y, fz)
        ax2 = fig.add_axes([0.42, 0.65, 0.32, 0.31])
        _plot_autovectors(spectra.index, u, ax2)
        ax3 = fig.add_axes([0.80, 0.65, 0.18, 0.31])
        plot_autovalues(s[:], ax3)  # don't plot last auto-values?
        ax5 = fig.add_axes([0.63, 0.08, 0.35, 0.50])
        plot_pca(v, ax5, x, hue_norm, palette)
        ylabel = "First Principal Component"
        ylabel_color = COLOR_MAP(0)
    else:  # Band integration
        ini, fin = band
        if ini not in spectra.index and fin not in spectra.index:
            msg = f"Band parameters ({ini}, {fin}) are not in the spectra's index."
            raise ValueError(msg)
        # columns index name are not necessarily unique
        y = np.array(
            [spectra.iloc[:, i].loc[ini:fin].sum() for i in range(spectra.shape[1])]
        )
        # rescale y
        y /= np.abs(y).max() / 10
        result = fit_binding(x, y, fz)
        ylabel = "Integrated Band Fluorescence"
        ylabel_color = "k"
    ax4 = fig.add_axes([0.05, 0.08, 0.50, 0.50])
    plot_lmfit(ax4, result, hue_norm, palette)
    _apply_common_plot_style(ax4, "LM fit", kind, "")
    ax4.set_ylabel(ylabel, color=ylabel_color)
    return fig, result


@typing.overload
def _binding_residuals(
    params: lmfit.Parameter,
    model_func: typing.Callable[
        [NDArray[np.float_], float, float, float], NDArray[np.float_]
    ],
    x: dict[str, NDArray[np.float_]],
    datasets: dict[str, NDArray[np.float_]],
) -> list[NDArray[np.float_]]:
    ...


@typing.overload
def _binding_residuals(
    params: lmfit.Parameter,
    model_func: typing.Callable[
        [NDArray[np.float_], float, float, float], NDArray[np.float_]
    ],
    x: dict[str, NDArray[np.float_]],
) -> dict[str, NDArray[np.float_]]:
    ...


def _binding_residuals(
    params: lmfit.Parameter,
    model_func: typing.Callable[
        [NDArray[np.float_], float, float, float], NDArray[np.float_]
    ],
    x: dict[str, NDArray[np.float_]],
    datasets: dict[str, NDArray[np.float_]] | None = None,
) -> dict[str, NDArray[np.float_]] | list[NDArray[np.float_]]:
    """Compute residuals for multiple datasets given a shared binding model function.

    Parameters
    ----------
    params : lmfit.Parameter
        Parameters for the model function.
    model_func : callable
        The model function used for fitting the data.
    x : list of np.ndarray
        Independent variables for each dataset.
    datasets : list of np.ndarray, optional
        The datasets to fit. If None, returns only the model predictions.

    Returns
    -------
    list of np.ndarray
        The residuals for each dataset or the model predictions if datasets is None.
    """
    models = {}
    for label in x:
        models[label] = model_func(
            x[label], params["K"], params[f"S0_{label}"], params[f"S1_{label}"]
        )
    # If no datasets are provided, return the model predictions
    if datasets is None:
        return models
    residuals = [datasets[label] - models[label] for label in x]
    return residuals



"""Command-line interface."""
from __future__ import annotations

import pprint
import warnings
from collections import namedtuple
from pathlib import Path

import click
import lmfit  # type: ignore
import numpy as np
import pandas as pd

from clophfit import __default_enspire_out_dir__
from clophfit import binding
from clophfit import prenspire
from clophfit import prtecan
from clophfit.prenspire import EnspireFile


def fit_routine(  # noqa: PLR0913
    titan: prtecan.TitrationAnalysis,
    kind: str,
    weight: bool,
    confint: float,
    norm: bool,
    bg: bool,
    dil: bool,
    verbose: int,
    out: Path,
    klim: tuple[float, float] | None,
    title: str | None,
    sel: tuple[float, float] | None,
    pdf: bool,
) -> None:
    """Help main."""
    titan.fit(
        kind,
        no_weight=(not weight),
        tval=float(confint),
        nrm=norm,
        bg=bg,
        dil=dil,
    )
    for i, fit in enumerate(titan.fittings):
        # Printout
        if verbose:
            try:
                meta = titan.labelblocksgroups[i].metadata
                print("-" * 79)
                print(f"\nlabel{i:d}")
                pprint.pprint(meta)
            except IndexError:
                print("-" * 79)
                print("\nGlobal on both labels")
            titan.print_fitting(i)
        # Csv tables
        fit.sort_index().to_csv(out / Path("ffit" + str(i) + ".csv"))
        if "SA2" in fit:
            out_cols = ["K", "sK", "SA", "sSA", "SB", "sSB"]
            out_cols.extend(["SA2", "sSA2", "SB2", "sSB2"])
        else:
            out_cols = ["K", "sK", "SA", "sSA", "SB", "sSB"]
        fit[out_cols].sort_index().to_csv(
            out / Path("fit" + str(i) + ".csv"), float_format="%5.1f"
        )
        # Plots
        f = titan.plot_k(i, xlim=klim, title=title)
        f.savefig(out / Path("K" + str(i) + ".png"))
        f = titan.plot_ebar(i, title=title)
        f.savefig(out / Path("ebar" + str(i) + ".png"))
        if sel:
            if kind.lower() == "ph":
                xmin, ymin = sel
                f = titan.plot_ebar(i, xmin=xmin, ymin=ymin, title=title)
            if kind.lower() == "cl":
                xmax, ymin = sel
                f = titan.plot_ebar(i, xmax=xmax, ymin=ymin, title=title)
            f.savefig(out / Path("ebarZ" + str(i) + ".png"))
    if pdf:
        titan.plot_all_wells(out / "all_wells.pdf")


@click.group()
@click.version_option(message="%(version)s")
def clop() -> None:  # pragma: no cover
    """Group command."""


@clop.command("eq1")
@click.argument("kd1", type=float)
@click.argument("pka", type=float)
@click.argument("ph", type=float)
def eq1(  # type: ignore
    kd1,
    pka,
    ph,
):
    """pH-deps for Kd."""
    print(binding.kd(kd1=kd1, pka=pka, ph=ph))


@clop.command("pr.tecan")
@click.argument("list_file", type=click.Path(path_type=Path))
@click.option(
    "--dil",
    type=click.Path(path_type=Path),
    is_flag=False,
    flag_value="additions.pH",
    show_default="additions.pH",
    help="Initial volume and additions",
)
@click.option("--norm", is_flag=True, help="Normalize using metadata (gain, flashes).")
@click.option(
    "--scheme",
    type=click.Path(path_type=Path),
    is_flag=False,
    flag_value="scheme.txt",
    show_default="scheme.txt",
    help="Positions of buffer and controls wells.",
)
@click.option("--bg", is_flag=True, help="Subtract buffer (scheme wells=='buffer').")
@click.option(
    "--kind",
    "-k",
    type=click.Choice(["pH", "Cl"], case_sensitive=False),
    default="pH",
    help="Kind of titration.",
    show_default=True,
)
@click.option(
    "--weight/--no-weight",
    default=False,
    show_default=True,
    help="Global fitting without relative residues weights.",
)
@click.option(
    "--fit/--no-fit",
    default=True,
    show_default=True,
    help="Perform also fit.",
)
@click.option("--fit-all", is_flag=True, help="Fit all exported data.")
@click.option(
    "--confint",
    default=0.95,
    type=click.FloatRange(0, 1, clamp=True),
    show_default=True,
    help="Confidence value for the calculation of parameter errors.",
)
@click.option(
    "--out",
    type=click.Path(path_type=Path),
    default="out2",
    help="Path to output results.",
    show_default=True,
)
@click.option("--pdf", is_flag=True, help="Full report in pdf file.")
@click.option("--title", "-t", default=None, help="Title for some plots.")
@click.option(
    "--Klim",
    default=None,
    type=(float, float),
    help="Range MIN, MAX (xlim) for plot_K.",
)
@click.option(
    "--sel",
    default=None,
    type=(float, float),
    help="Errorbar plot for selection with K_min SA_min.",
)
@click.option("--verbose", "-v", count=True, help="Verbosity of messages.")
def tecan(  # type: ignore  # noqa: PLR0913
    list_file,
    scheme,
    dil,
    verbose,
    bg,
    kind,
    norm,
    out,
    weight,
    fit,
    fit_all,
    confint,
    klim,
    title,
    sel,
    pdf,
):
    """Convert a list of plate reader acquisitions into titrations.

    LIST_FILE : List of Tecan files and concentration values.

    Save titrations as .dat files.

    Fits all wells using 2 labels and produces:

    - K plot

    - ebar and (for selection) ebarZ plot

    - all_wells pdf

    - csv tables for all labelblocks and global fittings.

    Notes: Buffer is always subtracted if scheme indicates buffer well positions.

    """
    titan = prtecan.TitrationAnalysis.fromlistfile(list_file)
    if scheme:
        titan.load_scheme(scheme)
        if dil:
            titan.load_additions(dil)
            bg = True  # should not be needed but influence decimals of
            # exported values; however ``dil imply bg```
            if kind.lower() == "cl":  # XXX cl conc must be elsewhere
                titan.conc = list(prtecan.calculate_conc(titan.additions, 1000.0))  # type: ignore
    titan.export_data(out)
    with (out / "metadata-labels.txt").open("w", encoding="utf-8") as fp:
        for lbg in titan.labelblocksgroups:
            pprint.pprint(lbg.metadata, stream=fp)
    if scheme:
        f = titan.plot_buffer(title=title)
        f.savefig(out / Path("buffer.png"))

    if fit:
        if bg and not scheme:
            # ``as bg requires scheme even though scheme does not imply bg```
            warnings.warn("Scheme is needed to compute buffer bg!", stacklevel=2)
        if fit_all:
            for n, b, d, out2 in [
                (0, 0, 0, "dat"),
                (1, 0, 0, "dat_nrm"),
                (0, 1, 0, "dat_bg"),
                (1, 1, 0, "dat_bg_nrm"),
                (0, 1, 1, "dat_bg_dil"),
                (1, 1, 1, "dat_bg_dil_nrm"),
            ]:
                out_fit = out / out2 / "0fit"
                out_fit.mkdir(parents=True, exist_ok=True)
                fit_routine(
                    titan,
                    kind,
                    weight,
                    confint,
                    bool(n),
                    bool(b),
                    bool(d),
                    verbose,
                    out_fit,
                    klim,
                    title,
                    sel,
                    pdf,
                )
        else:
            fit_routine(
                titan,
                kind,
                weight,
                confint,
                norm,
                bg,
                bool(dil),
                verbose,
                out,
                klim,
                title,
                sel,
                pdf,
            )


@clop.command("pr.enspire")
@click.argument("csv", type=click.Path(exists=True, path_type=Path))
@click.argument("note_fp", type=click.Path(exists=True, path_type=Path), required=False)
@click.option(
    "-b",
    "--band-intervals",
    "bands",
    multiple=True,
    nargs=3,
    type=(str, int, int),
    help="Label and band interval (format: LABEL LOWER UPPER)",
)
@click.option(
    "--out",
    "-d",
    type=click.Path(path_type=Path),
    default=__default_enspire_out_dir__,
    help="Path to output results.",
    show_default=True,
)
@click.option("--verbose", "-v", count=True, help="Verbosity of messages.")
def enspire(csv, note_fp, out, bands, verbose):  # type: ignore
    """Save spectra as csv tables from EnSpire xls file."""
    ef = EnspireFile(csv, verbose=verbose)
    ef.export_measurements(out)
    if note_fp is not None:
        fit_enspire(ef, note_fp, out, bands, verbose)


def fit_enspire(
    ef: EnspireFile,
    note_fp: Path,
    out_dir: Path,
    bands: list[tuple[str, int, int]] | None,
    verbose: int,
) -> None:
    """Fit prenspire titration (all labels, temp, mutant, titrations)."""
    note = prenspire.prenspire.Note(note_fp, verbose=verbose)
    note.build_titrations(ef)
    dbands = {label: (ini, fin) for label, ini, fin in bands} if bands else {}
    x_combined = {}
    y_combined = {}
    for name, d_name in note.titrations.items():
        for temp, d_temp in d_name.items():
            for tit, d_tit in d_temp.items():
                if tit.split("_")[0] == "pH":
                    ttype = "Cl"
                elif tit.split("_")[0] == "Cl":
                    ttype = "pH"
                else:
                    msg = "Unknown titration type."
                    raise ValueError(msg)
                for label, data in d_tit.items():
                    band: tuple[int, int] | None = dbands.get(label)
                    figure, result = binding.fitting.analyze_spectra(data, ttype, band)
                    if band:
                        x_combined[label] = result.userkws["x"]
                        y_combined[label] = result.data
                    pdf_file = out_dir / f"{name}_{temp}_{label}_{tit}_{band}.pdf"
                    figure.savefig(pdf_file)
                    _print_result(result, pdf_file, str(band))
                if len(d_tit.keys() - dbands.keys()) > 1 or len(dbands.keys()) > 1:
                    figs_res = binding.fitting.analyze_spectra_glob(
                        d_tit, ttype, dbands, x_combined, y_combined
                    )
                    figure_svd, result_svd, figure_bands, result_bands = figs_res
                    if figure_svd:
                        pdf_file = out_dir / f"{name}_{temp}_all_{tit}_SVD.pdf"
                        figure_svd.savefig(pdf_file)
                        _print_result(result_svd, pdf_file, "")
                    if figure_bands:
                        bands_slist = [f"{k}({v[0]},{v[1]})" for k, v in dbands.items()]
                        bands_str = "".join(bands_slist)
                        pdf_file = out_dir / f"{name}_{temp}_all_{tit}_{bands_str}.pdf"
                        figure_bands.savefig(pdf_file)
                        _print_result(result_bands, pdf_file, bands_str)


def _print_result(
    result: lmfit.model.ModelResult | lmfit.model.MinimizerResult,
    pdf_file: Path,
    band_str: str,
) -> None:
    print(str(pdf_file))
    print(f"Best fit using '{band_str}' band:\n")
    try:
        print(result.ci_report(ndigits=2, with_offset=False))
    except (ValueError, AttributeError):
        print(result.params)
    print(f"\n Plot saved in '{pdf_file}'.\n")


@clop.command("fit_titration")
@click.argument("csv_fp", type=click.Path(exists=True, path_type=Path))
@click.argument("note_fp", type=click.Path(exists=True, path_type=Path))
@click.option("-d", "--out", default=Path("."), type=Path, help="destination directory")
@click.option(
    "-t",
    "titration_type",
    default="pH",
    type=click.Choice(["pH", "Cl"], case_sensitive=False),
    help="titration type (default: pH)",
)
@click.option(
    "-b", "--band", nargs=2, type=int, help="Integration interval from <1> to <2>"
)
@click.option("-v", "--verbose", is_flag=True, help="increase output verbosity")
def fit_titration(csv_fp, note_fp, out, titration_type, band, verbose):  # type: ignore # noqa: PLR0913
    """Update old svd or band fit of titration spectra."""
    note_df = pd.read_csv(note_fp, sep="\t")
    csv = pd.read_csv(csv_fp)
    out_fp = Path(out)
    # Ignore buffer wells! SVD will use differences between spectra.
    note_df = note_df[note_df["mutant"] != "buffer"]
    Notes = namedtuple("Notes", ["wells", "conc"])
    note = Notes(list(note_df["well"]), list(note_df[titration_type]))
    spectra = csv[note.wells]
    spectra.index = csv["lambda"]
    spectra.columns = np.array(note.conc)
    if verbose:
        print(csv)
        click.echo(note_fp)
        print(note)
        print("DataFrame\n", spectra)
    figure, result = binding.fitting.analyze_spectra(spectra, titration_type, band)
    # output
    out_fp.mkdir(parents=True, exist_ok=True)
    pdf_file = out_fp / f"{csv_fp.stem}_{band}_{note_fp.stem}.pdf"
    figure.savefig(pdf_file)
    _print_result(result, pdf_file, str(band))



"""prtecan/prtecan.py."""
from __future__ import annotations

import hashlib
import itertools
import typing
import warnings
from contextlib import suppress
from dataclasses import InitVar
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import Sequence

import matplotlib.pyplot as plt  # type: ignore
import numpy as np
import pandas as pd
import seaborn as sb  # type: ignore  # noqa: ICN001
from matplotlib.backends.backend_pdf import PdfPages  # type: ignore
from numpy.typing import NDArray

from clophfit.binding import fit_titration
from clophfit.binding import fz_kd_singlesite
from clophfit.binding import fz_pk_singlesite

# list_of_lines
# after set([type(x) for l in csvl for x in l]) = float | int | str
DAT = "dat"
DAT_NRM = "dat_nrm"
DAT_BG = "dat_bg"
DAT_BG_NRM = "dat_bg_nrm"
DAT_BG_DIL = "dat_bg_dil"
DAT_BG_DIL_NRM = "dat_bg_dil_nrm"
STD_MD_LINE_LENGTH = 2
NUM_ROWS_96WELL = 8


def read_xls(path: Path) -> list[list[str | int | float]]:
    """Read first sheet of an xls file.

    Parameters
    ----------
    path : Path
        Path to `.xls` file.

    Returns
    -------
    list_of_lines
        Lines.

    """
    sheet = pd.read_excel(path)
    n0 = pd.DataFrame([[np.nan] * len(sheet.columns)], columns=sheet.columns)
    sheet = pd.concat([n0, sheet], ignore_index=True)
    sheet = sheet.fillna("")
    return list(sheet.to_numpy().tolist())


@typing.overload
def lookup_listoflines(csvl: list[list[str]], pattern: str, col: int) -> list[int]:
    ...


@typing.overload
def lookup_listoflines(
    csvl: list[list[str | int | float]], pattern: str = "Label: Label", col: int = 0
) -> list[int]:
    ...


def lookup_listoflines(
    csvl: list[list[str | int | float]] | list[list[str]],
    pattern: str = "Label: Label",
    col: int = 0,
) -> list[int]:
    """Lookup line numbers (row index) where given pattern occurs.

    Parameters
    ----------
    csvl : list_of_lines
        Lines of a csv/xls file.
    pattern : str
        Pattern to be searched (default="Label: Label").
    col : int
        Column to search (default=0).

    Returns
    -------
    list[int]
        Row/line index for all occurrences of pattern. Empty list for no occurrences.

    """
    indexes = []
    for i, line in enumerate(csvl):
        if isinstance(line, list) and col < len(line):
            element = line[col]
            if isinstance(element, (str, int, float)) and pattern in str(element):
                indexes.append(i)
    return indexes


def strip_lines(lines: list[list[str | int | float]]) -> list[list[str | int | float]]:
    """Remove empty fields/cells from lines read from a csv file.

    Parameters
    ----------
    lines : list_of_lines
        Lines that are a list of fields, typically from a csv/xls file.

    Returns
    -------
    list_of_lines
        Lines removed from blank cells.

    Examples
    --------
    >>> lines = [['Shaking (Linear) Amplitude:', '', '', '', 2, 'mm', '', '', '', '', '']]
    >>> strip_lines(lines)
    [['Shaking (Linear) Amplitude:', 2, 'mm']]

    """
    return [[e for e in line if e] for line in lines]


# TODO with a filter ectract_metadata with a map


@dataclass(frozen=False)
class Metadata:
    """Represents the value of a metadata dictionary.

    Parameters
    ----------
    value : int | str | float | None
        The value for the dictionary key.
    unit : Sequence[str | float | int] | None, optional
        The first element represents the unit, while the following elements are only listed.
    """

    value: int | str | float | None
    unit: Sequence[str | float | int] | None = None


def extract_metadata(
    lines: list[list[str | int | float]],
) -> dict[str, Metadata]:
    """Extract metadata into both Tecanfile and Labelblock.

    From a list of stripped lines takes the first field as the **key** of the
    metadata dictionary, remaining fields goes into a list of values with the
    exception of Label ([str]) and Temperature ([float]).

    Parameters
    ----------
    lines : list_of_lines
        Lines that are a list of fields, typically from a csv/xls file.

    Returns
    -------
    dict[str, str | list[str | int | float]]
        Metadata for Tecanfile or Labelblock.

    Examples
    --------
    >>> lines = [['Shaking (Linear) Amplitude:', '', '', '', 2, 'mm', '', '', '', '', '']]
    >>> extract_metadata(lines)
    {'Shaking (Linear) Amplitude:': Metadata(value=2, unit=['mm'])}

    >>> lines = [['', 'Temperature: 26 °C', '', '', '', '', '', '', '', '', '']]
    >>> extract_metadata(lines)
    {'Temperature': Metadata(value=26.0, unit=['°C'])}

    >>> lines = [['Excitation Wavelength', '', '', '', 400, 'nm', '', '', '', '', '']]
    >>> extract_metadata(lines)
    {'Excitation Wavelength': Metadata(value=400, unit=['nm'])}

    >>> lines = [['Label: Label1', '', '', '', '', '', '', '', '', '', '', '', '']]
    >>> extract_metadata(lines)
    {'Label': Metadata(value='Label1', unit=None)}

    >>> lines = [['Mode', '', '', '', 'Fluorescence Top Reading', '', '', '', '', '']]
    >>> extract_metadata(lines)['Mode'].value
    'Fluorescence Top Reading'

    """
    md: dict[str, Metadata] = {}

    for line in strip_lines(lines):
        if len(line) > STD_MD_LINE_LENGTH:
            md.update({str(line[0]): Metadata(line[1], line[2:])})
        elif len(line) == STD_MD_LINE_LENGTH:
            md.update({str(line[0]): Metadata(line[1])})
        elif len(line) == 1 and isinstance(line[0], str) and ":" in line[0]:
            k, v = line[0].split(":")
            vals: list[str] = v.split()
            val: float | str
            try:
                val = float(vals[0])
            except ValueError:
                val = vals[0]
            if len(vals) == 1:
                md.update({k: Metadata(val)})
            else:
                md.update({k: Metadata(val, vals[1:])})
        elif line:
            md.update({str(line[0]): Metadata(line[0])})

    return md


def merge_md(mds: list[dict[str, Metadata]]) -> dict[str, Metadata]:
    """Merge a list of metadata dict if the key value is the same in the list."""
    mmd = {k: v for k, v in mds[0].items() if all(v == md[k] for md in mds[1:])}
    # To account for the case 93"Optimal" and 93"Manual" in lb metadata
    if mmd.get("Gain") is None and mds[0].get("Gain") is not None:  # noqa: SIM102
        if all(mds[0]["Gain"].value == md["Gain"].value for md in mds[1:]):
            mmd["Gain"] = Metadata(mds[0]["Gain"].value)
    return mmd


def calculate_conc(
    additions: Sequence[float], conc_stock: float, conc_ini: float = 0.0
) -> NDArray[np.float_]:
    """Calculate concentration values.

    additions[0]=vol_ini; Stock concentration is a parameter.

    Parameters
    ----------
    additions : Sequence[float]
        Initial volume and all subsequent additions.
    conc_stock : float
        Concentration of the stock used for additions.
    conc_ini : float
        Initial concentration (default=0).

    Returns
    -------
    np.ndarray
        Concentrations as vector.

    """
    vol_tot = np.cumsum(additions)
    concs = np.ones(len(additions))
    concs[0] = conc_ini
    for i, add in enumerate(additions[1:], start=1):
        concs[i] = (concs[i - 1] * vol_tot[i - 1] + conc_stock * float(add)) / vol_tot[
            i
        ]
    return concs  # , vol_tot


def dilution_correction(additions: list[float]) -> NDArray[np.float_]:
    """Apply dilution correction.

    Parameters
    ----------
    additions: list[float]
        List of initial volume (index=0) followed by all additions.

    Returns
    -------
    NDArray[float]
        Dilution correction vector.
    """
    volumes = np.cumsum(additions)
    corrections: NDArray[np.float_] = volumes / volumes[0]
    return corrections


class BufferWellsMixin:
    """Mixin class for handling buffer wells.

    This mixin adds a property for buffer wells. It's intended to be used in
    classes that deal with collections of wells, where some of the wells are
    designated as buffer wells.
    """

    _buffer_wells: list[str] | None = None

    @property
    def buffer_wells(self) -> list[str] | None:
        """List of buffer wells."""
        return self._buffer_wells

    @buffer_wells.setter
    def buffer_wells(self, value: list[str]) -> None:
        self._buffer_wells = value
        self.on_buffer_wells_set(value)

    def on_buffer_wells_set(self, value: list[str]) -> None:
        """Provide a hook for subclasses to add behavior when buffer_wells is set."""
        pass


@dataclass
class Labelblock(BufferWellsMixin):
    """Parse a label block.

    Parameters
    ----------
    _lines : list[list[str | int | float]]
        Lines to create this Labelblock.
    path : Path, optional
        Path to the containing file, if it exists.

    Raises
    ------
    Exception
        When data do not correspond to a complete 96-well plate.

    Warns
    -----
    Warning
        When it replaces "OVER" with ``np.nan`` for saturated values.
    """

    _lines: InitVar[list[list[str | int | float]]]
    path: Path | None = None
    #: Metadata specific for this Labelblock.
    metadata: dict[str, Metadata] = field(init=False, repr=True)
    #: The 96 data values as {'well_name', value}.
    data: dict[str, float] = field(init=False, repr=True)
    _KEYS: typing.ClassVar[list[str]] = [
        "Emission Bandwidth",
        "Emission Wavelength",
        "Excitation Bandwidth",
        "Excitation Wavelength",
        "Mode",
        "Integration Time",
        "Number of Flashes",
    ]
    _NORM_KEYS: typing.ClassVar[list[str]] = [
        "Integration Time",
        "Number of Flashes",
        "Gain",
    ]

    def __post_init__(self, lines: list[list[str | int | float]]) -> None:
        """Create metadata and data; initialize labelblock's properties."""
        self._buffer_wells: list[str] | None = None
        self._buffer: float | None = None
        self._buffer_norm: float | None = None
        self._buffer_sd: float | None = None
        self._buffer_norm_sd: float | None = None
        self._data_norm: dict[str, float] | None = None
        self._data_buffersubtracted: dict[str, float] | None = None
        self._data_buffersubtracted_norm: dict[str, float] | None = None
        if lines[14][0] == "<>" and lines[23] == lines[24] == [""] * 13:
            stripped = strip_lines(lines)
            stripped[14:23] = []
            self.metadata: dict[str, Metadata] = extract_metadata(stripped)
            self.data: dict[str, float] = self._extract_data(lines[15:23])
        else:
            msg = "Cannot build Labelblock: not 96 wells?"
            raise ValueError(msg)

    def _extract_data(self, lines: list[list[str | int | float]]) -> dict[str, float]:
        """Extract data into a dictionary.

        {'A01' : value}
        :
        {'H12' : value}

        Parameters
        ----------
        lines : list_of_lines
            xls file read into lines.

        Returns
        -------
        dict[str, float]
            Data from a label block.

        Raises
        ------
        ValueError
            When something went wrong. Possibly because not 96-well.

        Warns
        -----
            When a cell contains saturated signal (converted into np.nan).

        """
        rownames = tuple("ABCDEFGH")
        data = {}
        try:
            assert len(lines) == NUM_ROWS_96WELL  # noqa: S101 # nosec B101
            for i, row in enumerate(rownames):
                assert lines[i][0] == row  # nosec B101 # noqa: S101 # e.g. "A" == "A"
                for col in range(1, 13):
                    try:
                        data[row + f"{col:0>2}"] = float(lines[i][col])
                    except ValueError:
                        data[row + f"{col:0>2}"] = np.nan
                        warnings.warn(
                            f"OVER\n Overvalue in {self.metadata['Label'].value}:"
                            f"{row}{col:0>2} of tecanfile {self.path}",
                            stacklevel=2,
                        )
        except AssertionError as exc:
            msg = "Cannot extract data in Labelblock: not 96 wells?"
            raise ValueError(msg) from exc
        return data

    def on_buffer_wells_set(self, value: list[str]) -> None:
        """Update related attributes upon setting 'buffer_wells' in Labelblock class.

        Parameters
        ----------
        value: list[str]
            The new value of 'buffer_wells'
        """
        self._buffer = float(np.average([self.data[k] for k in value]))
        self._buffer_sd = float(np.std([self.data[k] for k in value]))
        self._buffer_norm = float(np.average([self.data_norm[k] for k in value]))
        self._buffer_norm_sd = float(np.std([self.data_norm[k] for k in value]))
        self._data_buffersubtracted = None
        self._data_buffersubtracted_norm = None

    @property
    def buffer(self) -> float | None:
        """Background value to be subtracted before dilution correction."""
        return self._buffer

    @buffer.setter
    def buffer(self, value: float) -> None:
        if self._buffer == value:
            return None
        self._data_buffersubtracted = None
        self._buffer = value

    @property
    def buffer_norm(self) -> float | None:
        """Background value to be subtracted before dilution correction."""
        return self._buffer_norm

    @buffer_norm.setter
    def buffer_norm(self, value: float) -> None:
        if self._buffer_norm == value:
            return None
        self._data_buffersubtracted_norm = None
        self._buffer_norm = value

    @property
    def buffer_sd(self) -> float | None:
        """Get standard deviation of buffer_wells values."""
        return self._buffer_sd

    @property
    def buffer_norm_sd(self) -> float | None:
        """Get standard deviation of normalized buffer_wells values."""
        return self._buffer_norm_sd

    @property
    def data_norm(self) -> dict[str, float]:
        """Normalize data by number of flashes, integration time and gain."""
        if self._data_norm is None:
            try:
                norm = 1000.0
                for k in Labelblock._NORM_KEYS:
                    norm /= self.metadata[k].value  # type: ignore # XXX: D
            except TypeError:
                warnings.warn(
                    "Could not normalize for non numerical Gain, Number of Flashes or Integration time.",
                    stacklevel=2,
                )  # pragma: no cover
            self._data_norm = {k: v * norm for k, v in self.data.items()}
        return self._data_norm

    @property
    def data_buffersubtracted(self) -> dict[str, float]:
        """Buffer subtracted data."""
        if self._data_buffersubtracted is None:
            if self.buffer:
                self._data_buffersubtracted = {
                    k: v - self.buffer for k, v in self.data.items()
                }
            else:
                self._data_buffersubtracted = {}
        return self._data_buffersubtracted

    @property
    def data_buffersubtracted_norm(self) -> dict[str, float]:
        """Normalize buffer-subtracted data."""
        if self._data_buffersubtracted_norm is None:
            if self.buffer_norm:
                self._data_buffersubtracted_norm = {
                    k: v - self.buffer_norm for k, v in self.data_norm.items()
                }
            else:
                self._data_buffersubtracted_norm = {}
        return self._data_buffersubtracted_norm

    def __eq__(self, other: object) -> bool:
        """Two labelblocks are equal when metadata KEYS are identical."""
        if not isinstance(other, Labelblock):
            return NotImplemented
        eq: bool = True
        for k in Labelblock._KEYS:
            eq &= self.metadata[k] == other.metadata[k]
        # 'Gain': [81.0, 'Manual'] = 'Gain': [81.0, 'Optimal'] They are equal
        eq &= self.metadata["Gain"].value == other.metadata["Gain"].value
        return eq

    def __almost_eq__(self, other: Labelblock) -> bool:
        """Two labelblocks are almost equal when they could be merged after normalization."""
        eq: bool = True
        # Integration Time, Number of Flashes and Gain can differ.
        for k in Labelblock._KEYS[:5]:
            eq &= self.metadata[k] == other.metadata[k]
        return eq


@dataclass
class Tecanfile:
    """Parse a Tecan .xls file.

    Parameters
    ----------
    path: Path
        Path to `.xls` file.

    Raises
    ------
    FileNotFoundError
        When path does not exist.
    Exception
        When no Labelblock is found.
    """

    path: Path
    #: General metadata for Tecanfile, like `Date` and `Shaking Duration`.
    metadata: dict[str, Metadata] = field(init=False, repr=True)
    #: All labelblocks contained in this file.
    labelblocks: list[Labelblock] = field(init=False, repr=True)

    def __post_init__(self) -> None:
        """Initialize."""
        csvl = read_xls(self.path)
        idxs = lookup_listoflines(csvl, pattern="Label: Label", col=0)
        if len(idxs) == 0:
            msg = "No Labelblock found."
            raise ValueError(msg)
        self.metadata = extract_metadata(csvl[: idxs[0]])
        labelblocks = []
        n_labelblocks = len(idxs)
        idxs.append(len(csvl))
        for i in range(n_labelblocks):
            labelblocks.append(Labelblock(csvl[idxs[i] : idxs[i + 1]], self.path))
        if any(
            labelblocks[i] == labelblocks[j]
            for i, j in itertools.combinations(range(n_labelblocks), 2)
        ):
            warnings.warn("Repeated labelblocks", stacklevel=2)
        self.labelblocks = labelblocks


@dataclass
class LabelblocksGroup(BufferWellsMixin):
    """Group labelblocks with compatible metadata.

    `data_norm` always exist.

    Parameters
    ----------
    labelblocks: list[Labelblock]
        Labelblocks to be grouped.
    allequal: bool
        True if labelblocks already tested equal.

    Raises
    ------
    Exception
        When labelblocks are neither equal nor almost equal.
    """

    labelblocks: list[Labelblock]
    allequal: bool = False
    #: Metadata shared by all labelblocks.
    metadata: dict[str, Metadata] = field(init=False, repr=True)

    def __post_init__(self) -> None:
        """Create common metadata and data."""
        labelblocks = self.labelblocks
        allequal = self.allequal
        self._buffer_wells: list[str] | None = None
        self._data: dict[str, list[float]] | None = None
        self._data_norm: dict[str, list[float]] | None = None
        self._data_buffersubtracted: dict[str, list[float]] | None = None
        self._data_buffersubtracted_norm: dict[str, list[float]] | None = None
        if not allequal:
            allequal = all(labelblocks[0] == lb for lb in labelblocks[1:])
        if allequal:
            self._data = {}
            for key in labelblocks[0].data:
                self._data[key] = [lb.data[key] for lb in labelblocks]
        # labelblocks that can be merged only after normalization
        elif all(labelblocks[0].__almost_eq__(lb) for lb in labelblocks[1:]):
            self._data_norm = {}
            for key in labelblocks[0].data:
                self._data_norm[key] = [lb.data_norm[key] for lb in labelblocks]
        else:
            msg = "Creation of labelblock group failed."
            raise ValueError(msg)
        self.labelblocks = labelblocks
        self.metadata = merge_md([lb.metadata for lb in labelblocks])

    @property
    def data(self) -> dict[str, list[float]] | None:
        """Return None or data."""
        return self._data

    @property
    def data_norm(self) -> dict[str, list[float]]:
        """Normalize data by number of flashes, integration time and gain."""
        if self._data_norm is None:
            self._data_norm = {}
            for key in self.labelblocks[0].data:
                self._data_norm[key] = [lb.data_norm[key] for lb in self.labelblocks]
        return self._data_norm

    def on_buffer_wells_set(self, value: list[str]) -> None:
        """Update related attributes upon setting 'buffer_wells' in Labelblock class.

        Parameters
        ----------
        value: list[str]
            The new value of 'buffer_wells'

        """
        for lb in self.labelblocks:
            lb.buffer_wells = value
        self._data_buffersubtracted = None
        self._data_buffersubtracted_norm = None

    @property
    def data_buffersubtracted(self) -> dict[str, list[float]] | None:
        """Buffer subtracted data."""
        if self.data is None:
            return None
        if self._data_buffersubtracted is None:
            self._data_buffersubtracted = (
                {
                    key: [lb.data_buffersubtracted[key] for lb in self.labelblocks]
                    for key in self.labelblocks[0].data
                }
                if self.buffer_wells
                else {}
            )
        return self._data_buffersubtracted

    @property
    def data_buffersubtracted_norm(self) -> dict[str, list[float]]:
        """Buffer subtracted data."""
        if self._data_buffersubtracted_norm is None:
            self._data_buffersubtracted_norm = (
                {
                    key: [lb.data_buffersubtracted_norm[key] for lb in self.labelblocks]
                    for key in self.labelblocks[0].data
                }
                if self.buffer_wells
                else {}
            )
        return self._data_buffersubtracted_norm


@dataclass
class TecanfilesGroup:
    """Group of Tecanfiles containing at least one common Labelblock.

    Parameters
    ----------
    tecanfiles: list[Tecanfile]
        List of Tecanfiles.

    Raises
    ------
    Exception
        When all Labelblocks are not at least almost equal.

    Warns
    -----
    Warning
        The Tecanfiles listed in `tecanfiles` are expected to contain the
        "same" list (of length N) of Labelblocks. Normally, N labelblocksgroups
        will be created. However, if not all Tecanfiles contain the same number
        of Labelblocks that can be merged ('equal' mergeable) in the same order,
        then a warning will be raised. In this case, a number M < N of groups
        can be built.
    """

    tecanfiles: list[Tecanfile]

    #: Each group contains its own data like a titration. ??
    labelblocksgroups: list[LabelblocksGroup] = field(init=False, default_factory=list)
    #: Metadata shared by all tecanfiles.
    metadata: dict[str, Metadata] = field(init=False, repr=True)

    def __post_init__(self) -> None:
        """Create metadata and labelblocksgroups."""
        n_labelblocks = [len(tf.labelblocks) for tf in self.tecanfiles]
        tf0 = self.tecanfiles[0]
        if all(tf0.labelblocks == tf.labelblocks for tf in self.tecanfiles[1:]):
            # Same number and order of labelblocks
            for i, _lb in enumerate(tf0.labelblocks):
                self.labelblocksgroups.append(
                    LabelblocksGroup(
                        [tf.labelblocks[i] for tf in self.tecanfiles], allequal=True
                    )
                )
        else:
            # Create as many as possible groups of labelblocks
            rngs = tuple(range(n) for n in n_labelblocks)
            for idx in itertools.product(*rngs):
                try:
                    gr = LabelblocksGroup(
                        [tf.labelblocks[idx[i]] for i, tf in enumerate(self.tecanfiles)]
                    )
                except ValueError:
                    continue
                # if labelblocks are all 'equal'
                else:
                    self.labelblocksgroups.append(gr)
            files = [tf.path for tf in self.tecanfiles]
            if len(self.labelblocksgroups) == 0:  # == []
                msg = f"No common labelblock in filenames: {files}."
                raise ValueError(msg)
            warnings.warn(
                f"Different LabelblocksGroup among filenames: {files}.", stacklevel=2
            )
        self.metadata = merge_md([tf.metadata for tf in self.tecanfiles])


@dataclass
class Titration(TecanfilesGroup, BufferWellsMixin):
    """TecanfileGroup + concentrations.

    Parameters
    ----------
    tecanfiles : list[Tecanfile]
        Tecanfiles to be grouped.
    conc : Sequence[float]
        Concentration or pH values.
    """

    tecanfiles: list[Tecanfile]
    conc: Sequence[float]

    def __post_init__(self) -> None:
        """Set up the initial values for the properties."""
        super().__post_init__()
        self._additions: list[float] | None = None
        self._data_dilutioncorrected: list[dict[str, list[float]] | None] | None = None
        self._data_dilutioncorrected_norm: list[dict[str, list[float]]] | None = None
        self._buffer_wells: list[str] | None = None
        self._dil_corr: NDArray[np.float_] = field(init=False, repr=False)

    @classmethod
    def fromlistfile(cls, list_file: Path | str) -> Titration:
        """Build `Titration` from a list[.pH|.Cl] file.

        Parameters
        ----------
        list_file: Path | str
            File path to the listfile ([fpath conc]).

        Returns
        -------
        Titration
        """
        tecanfiles, conc = Titration._listfile(Path(list_file))
        return cls(tecanfiles, conc)

    @staticmethod
    def _listfile(listfile: Path) -> tuple[list[Tecanfile], Sequence[float]]:
        """Help construction from file.

        Parameters
        ----------
        listfile: Path
            File path to the listfile ([fpath conc]).

        Returns
        -------
        tecanfiles and conc.

        Raises
        ------
        FileNotFoundError
            When cannot access `list_file`.
        ValueError
            For unexpected file format, e.g. length of filename column differs from
            length of conc values.
        """
        try:
            table = pd.read_csv(listfile, sep="\t", names=["filenames", "conc"])
        except FileNotFoundError as exc:
            msg = f"Cannot find: {listfile}"
            raise FileNotFoundError(msg) from exc
        if table["filenames"].count() != table["conc"].count():
            msg = f"Check format [filenames conc] for listfile: {listfile}"
            raise ValueError(msg)
        conc = table["conc"].tolist()
        tecanfiles = [Tecanfile(listfile.parent / f) for f in table["filenames"]]
        return tecanfiles, conc

    @property
    def additions(self) -> list[float] | None:
        """List of initial volume followed by additions."""
        return self._additions

    # Here there is not any check on the validity of additions (e.g. length).
    @additions.setter
    def additions(self, additions: list[float]) -> None:
        self._additions = additions
        self._dil_corr = dilution_correction(additions)
        self._data_dilutioncorrected = None
        self._data_dilutioncorrected_norm = None

    def load_additions(self, additions_file: Path) -> None:
        """Load additions from file."""
        additions = pd.read_csv(additions_file, names=["add"])
        self.additions = additions["add"].tolist()

    def on_buffer_wells_set(self, value: list[str]) -> None:
        """Update related attributes upon setting 'buffer_wells' in Labelblock class.

        Parameters
        ----------
        value: list[str]
            The new value of 'buffer_wells'

        """
        for lbg in self.labelblocksgroups:
            lbg.buffer_wells = value
        self._data_dilutioncorrected = None
        self._data_dilutioncorrected_norm = None

    @property
    def data_dilutioncorrected(self) -> list[dict[str, list[float]] | None] | None:
        """Buffer subtracted data."""
        if self._data_dilutioncorrected is None and self.additions:
            self._data_dilutioncorrected = [
                {
                    k: (np.array(v) * self._dil_corr).tolist()
                    for k, v in lbg.data_buffersubtracted.items()
                }
                if lbg.data_buffersubtracted
                else None
                for lbg in self.labelblocksgroups
            ]
        return self._data_dilutioncorrected

    @property
    def data_dilutioncorrected_norm(self) -> list[dict[str, list[float]]] | None:
        """Buffer subtracted data."""
        if self._data_dilutioncorrected_norm is None and self.additions:
            self._data_dilutioncorrected_norm = [
                {
                    k: (np.array(v) * self._dil_corr).tolist()
                    for k, v in lbg.data_buffersubtracted_norm.items()
                }
                for lbg in self.labelblocksgroups
            ]
        return self._data_dilutioncorrected_norm

    def export_data(self, out_folder: Path) -> None:
        """Export dat files [x,y1,..,yN] from labelblocksgroups.

        Remember that a Titration has at least 1 normalized Lbg dataset `dat_nrm`.

        dat:            [d1, None] | [d1, d2]
        dat_bg:         [{}, None] | [d1, None] | [{}, {}] | [d1, d2]
        dat_bg_dil:     [{}, None] | [d1, None] | [{}, {}] | [d1, d2]
        dat_nrm:        [d1,d2]
        dat_bg_nrm:     [{}, {}] | [d1, d2]
        dat_bg_dil_nrm: [{}, {}] | [d1, d2]

        Parameters
        ----------
        out_folder : Path
            Path to output folder.

        """
        out_folder.mkdir(parents=True, exist_ok=True)

        def write(
            conc: Sequence[float], data: list[dict[str, list[float]]], out_folder: Path
        ) -> None:
            """Write data."""
            if any(data):
                out_folder.mkdir(parents=True, exist_ok=True)
                columns = ["x"] + [f"y{i}" for i in range(1, len(data) + 1)]
                for key in data[0]:
                    dat = np.vstack((conc, [dt[key] for dt in data]))
                    datxy = pd.DataFrame(dat.T, columns=columns)
                    datxy.to_csv(
                        out_folder / Path(key).with_suffix(".dat"), index=False
                    )

        write(
            self.conc,
            [lbg.data for lbg in self.labelblocksgroups if lbg.data],
            out_folder / DAT,
        )
        write(
            self.conc,
            [lbg.data_norm for lbg in self.labelblocksgroups],
            out_folder / DAT_NRM,
        )
        write(
            self.conc,
            [
                lbg.data_buffersubtracted
                for lbg in self.labelblocksgroups
                if lbg.data_buffersubtracted
            ],
            out_folder / DAT_BG,
        )
        write(
            self.conc,
            [lbg.data_buffersubtracted_norm for lbg in self.labelblocksgroups],
            out_folder / DAT_BG_NRM,
        )
        if self.data_dilutioncorrected:
            write(
                self.conc,
                [e for e in self.data_dilutioncorrected if e],
                out_folder / DAT_BG_DIL,
            )
        if self.data_dilutioncorrected_norm:
            write(
                self.conc,
                self.data_dilutioncorrected_norm,
                out_folder / DAT_BG_DIL_NRM,
            )


@dataclass
class PlateScheme:
    """Define buffer, ctrl and unk wells, and ctrl names.

    Parameters
    ----------
    file: Path
        File path to the scheme file [<well Id, sample name>].
    """

    file: Path | None = None
    _buffer: list[str] = field(default_factory=list, init=False)
    _ctrl: list[str] = field(default_factory=list, init=False)
    _names: dict[str, set[str]] = field(default_factory=dict, init=False)

    @property
    def buffer(self) -> list[str]:
        """List of buffer wells."""
        return self._buffer

    @buffer.setter
    def buffer(self, value: list[str]) -> None:
        if not all(isinstance(item, str) for item in value):
            msg = "Buffer wells must be a list of strings"
            raise TypeError(msg)
        self._buffer = value

    @property
    def ctrl(self) -> list[str]:
        """List of CTR wells."""
        return self._ctrl

    @ctrl.setter
    def ctrl(self, value: list[str]) -> None:
        if not all(isinstance(item, str) for item in value):
            msg = "Ctrl wells must be a list of strings"
            raise TypeError(msg)
        self._ctrl = value

    @property
    def names(self) -> dict[str, set[str]]:
        """A dictionary mapping sample names to their associated list of wells."""
        return self._names

    @names.setter
    def names(self, value: dict[str, set[str]]) -> None:
        msg = "Names must be a dictionary mapping strings to sets of strings"
        if not isinstance(value, dict):
            raise TypeError(msg)
        if not all(
            isinstance(k, str)  # type: ignore
            and isinstance(v, set)  # type: ignore
            and all(isinstance(item, str) for item in v)
            for k, v in value.items()
        ):
            raise TypeError(msg)
        self._names = value

    def __post_init__(self) -> None:
        """Complete initialization."""
        if self.file:
            table = pd.read_csv(self.file, sep="\t")
            if (
                table.columns.tolist() != ["well", "sample"]
                or table["well"].count() != table["sample"].count()
            ):
                msg = f"Check format [well sample] for schemefile: {self.file}"
                raise ValueError(msg)
            scheme = table.groupby("sample")["well"].unique()
            self.buffer = list(scheme["buffer"])
            self.ctrl = list(
                {well for sample in scheme.tolist() for well in sample}
                - set(self.buffer)
            )
            self.names = {str(k): set(v) for k, v in scheme.items() if k != "buffer"}


class FitFirstError(Exception):
    """Error when plotting before fitting."""

    def __init__(self) -> None:
        super().__init__("Run fit first")


@dataclass
class TitrationAnalysis(Titration):
    """Perform analysis of a titration.

    Raises
    ------
    ValueError
        For unexpected file format, e.g. header `names`.
    """

    #: List of result dataframes.
    fittings: list[pd.DataFrame] = field(init=False, default_factory=list)
    #: Function used in the fitting.
    fz: typing.Callable[
        [float, NDArray[np.float_] | Sequence[float], NDArray[np.float_]],
        NDArray[np.float_],
    ] = field(init=False)
    #: A list of wells containing samples that are neither buffer nor CTR samples.
    keys_unk: list[str] = field(init=False, default_factory=list)

    def __post_init__(self) -> None:
        """Set up the initial values for the properties."""
        super().__post_init__()
        self._scheme: PlateScheme = PlateScheme()
        self._datafit: Sequence[dict[str, list[float]] | None] = []

    @classmethod
    def fromlistfile(cls, list_file: Path | str) -> TitrationAnalysis:
        """Build `TitrationAnalysis` from a list[.pH|.Cl] file.

        Parameters
        ----------
        list_file: Path | str
            File path to the listfile ([fpath conc]).

        Returns
        -------
        TitrationAnalysis
        """
        tecanfiles, conc = TitrationAnalysis._listfile(Path(list_file))
        return cls(tecanfiles, conc)

    @property
    def scheme(self) -> PlateScheme:
        """Scheme for known samples like {'buffer', ['H12', 'H01'], 'ctrl'...}."""
        return self._scheme

    def load_scheme(self, schemefile: Path) -> None:
        """Load scheme from file. Set buffer_wells."""
        self._scheme = PlateScheme(schemefile)
        self.buffer_wells = self._scheme.buffer

    def fit(  # noqa: PLR0913, PLR0912
        self,
        kind: str,
        ini: int = 0,
        fin: int | None = None,
        no_weight: bool = False,
        tval: float = 0.95,
        nrm: bool = False,
        bg: bool = False,
        dil: bool = False,
    ) -> None:
        """Fit titrations.

        Here is less general. It is for 2 labelblocks.

        Parameters
        ----------
        kind : str
            Titration type {'pH'|'Cl'}
        ini : int
            Initial point (default: 0).
        fin : int, optional
            Final point (default: None).
        no_weight : bool
            Do not use residues from single Labelblock fit as weight for global fitting.
        tval : float
            Only for tval different from default=0.95 for the confint calculation.
        nrm: bool
            Data normalization flag (default=False).
        bg: bool
            Buffer subtraction flag (default=False).
        dil: bool
            Dilution correction flag (default=False).

        Notes
        -----
        Create (: list) 3 fitting tables into self.fittings.

        """
        if kind == "Cl":
            self.fz = fz_kd_singlesite
        elif kind == "pH":
            self.fz = fz_pk_singlesite
        x = np.array(self.conc)
        fittings = []
        # datafit
        if dil:
            if nrm and self.data_dilutioncorrected_norm:
                # maybe need also bool(any([{}, {}])) or np.sum([bool(e) for e in [{}, {}]])
                self._datafit = self.data_dilutioncorrected_norm
            elif self.data_dilutioncorrected:
                self._datafit = self.data_dilutioncorrected
            else:  # back up to dat_nrm
                warnings.warn(
                    "No dilution corrected data found; use normalized data.",
                    stacklevel=2,
                )
                self._datafit = [lbg.data_norm for lbg in self.labelblocksgroups]
        elif bg:
            if nrm:
                self._datafit = [
                    lbg.data_buffersubtracted_norm for lbg in self.labelblocksgroups
                ]
            else:
                self._datafit = [
                    lbg.data_buffersubtracted for lbg in self.labelblocksgroups
                ]
        elif nrm:
            self._datafit = [lbg.data_norm for lbg in self.labelblocksgroups]
        else:
            self._datafit = [lbg.data for lbg in self.labelblocksgroups]

        # Any Lbg at least contains normalized data.
        keys_fit = self.labelblocksgroups[0].data_norm.keys() - set(self.scheme.buffer)
        self.keys_unk = list(keys_fit - set(self.scheme.ctrl))

        for data in self._datafit:
            fitting = pd.DataFrame()
            if data:
                for k in keys_fit:
                    y = data[k]
                    res = fit_titration(
                        kind, self.conc[ini:fin], np.array(y[ini:fin]), tval_conf=tval
                    )
                    res.index = pd.Index([k])
                    # fitting = fitting.append(res, sort=False) DDD
                    fitting = pd.concat([fitting, res], sort=False)
                    # TODO assert (fitting.columns == res.columns).all()
                    # better to refactor this function
                fittings.append(fitting)
        # Global weighted on relative residues of single fittings.
        fitting = pd.DataFrame()
        for k in keys_fit:
            # Actually y or y2 can be None (because it was possible to build only 1 Lbg)
            y = self._datafit[0][k]  # type: ignore # XXX: D
            y2 = self._datafit[1][k]  # type: ignore # XXX: D
            residue = y - self.fz(
                fittings[0]["K"].loc[k],
                [fittings[0]["SA"].loc[k], fittings[0]["SB"].loc[k]],
                x,
            )
            residue /= y  # TODO residue or
            # log(residue/y) https://www.tandfonline.com/doi/abs/10.1080/00031305.1985.10479385
            residue2 = y2 - self.fz(
                fittings[1]["K"].loc[k],
                [fittings[1]["SA"].loc[k], fittings[1]["SB"].loc[k]],
                x,
            )
            residue2 /= y2
            if no_weight:
                for i, _rr in enumerate(residue):
                    residue[i] = 1  # TODO use np.ones() but first find a way to test
                    residue2[i] = 1
            res = fit_titration(
                kind,
                self.conc[ini:fin],
                np.array(y[ini:fin]),
                y2=np.array(y2[ini:fin]),
                residue=residue[ini:fin],
                residue2=residue2[ini:fin],
                tval_conf=tval,
            )
            res.index = pd.Index([k])
            # fitting = fitting.append(res, sort=False) DDD
            fitting = pd.concat([fitting, res], sort=False)
        fittings.append(fitting)
        # Write the name of the control e.g. S202N in the "ctrl" column
        for fitting in fittings:
            for ctrl_name, wells in self.scheme.names.items():
                for well in wells:
                    fitting.loc[well, "ctrl"] = ctrl_name
        self.fittings = fittings

    def plot_k(
        self,
        lb: int,
        xlim: tuple[float, float] | None = None,
        title: str | None = None,
    ) -> plt.figure:
        """Plot K values as stripplot.

        Parameters
        ----------
        lb: int
            Labelblock index.
        xlim : tuple[float, float], optional
            Range.
        title : str, optional
            To name the plot.

        Returns
        -------
        plt.figure
            The figure.

        Raises
        ------
        FitFirstError
            When no fitting results are yet available.

        """
        if not hasattr(self, "fittings"):
            raise FitFirstError()
        sb.set(style="whitegrid")
        f = plt.figure(figsize=(12, 16))
        # Ctrl
        ax1 = plt.subplot2grid((8, 1), loc=(0, 0))
        if len(self.scheme.ctrl) > 0:
            res_ctrl = self.fittings[lb].loc[self.scheme.ctrl].sort_values("ctrl")
            sb.stripplot(
                x=res_ctrl["K"],
                y=res_ctrl.index,
                size=8,
                orient="h",
                hue=res_ctrl.ctrl,
                ax=ax1,
            )
            plt.errorbar(
                res_ctrl.K,
                range(len(res_ctrl)),
                xerr=res_ctrl["sK"],  # xerr=res_ctrl.sK*res_ctrl.tval,
                fmt=".",
                c="lightgray",
                lw=8,
            )
            plt.grid(1, axis="both")
        # Unk
        res_unk = self.fittings[lb].loc[self.keys_unk].sort_index(ascending=False)
        ax2 = plt.subplot2grid((8, 1), loc=(1, 0), rowspan=7)
        sb.stripplot(
            x=res_unk["K"],
            y=res_unk.index,
            size=12,
            orient="h",
            lw=2,
            palette="Blues",
            hue=res_unk["SA"],
            ax=ax2,
        )
        plt.legend("")
        plt.errorbar(
            res_unk["K"],
            range(len(res_unk)),
            xerr=res_unk["sK"],
            fmt=".",
            c="gray",
            lw=2,
        )
        plt.yticks(range(len(res_unk)), res_unk.index)
        plt.ylim(-1, len(res_unk))
        plt.grid(1, axis="both")
        if not xlim:
            xlim = (res_unk["K"].min(), res_unk["K"].max())
            if len(self.scheme.ctrl) > 0:
                xlim = (
                    0.99 * min(res_ctrl["K"].min(), xlim[0]),
                    1.01 * max(res_ctrl["K"].max(), xlim[1]),
                )
            xlim = (0.99 * xlim[0], 1.01 * xlim[1])
        ax1.set_xlim(xlim)
        ax2.set_xlim(xlim)
        ax1.set_xticklabels([])
        ax1.set_xlabel("")
        title = title if title else ""
        title += "  label:" + str(lb)
        f.suptitle(title, fontsize=16)
        f.tight_layout(pad=1.2, w_pad=0.1, h_pad=0.5, rect=(0, 0, 1, 0.97))
        return f

    def plot_well(self, key: str) -> plt.figure:
        """Plot global fitting using 2 labelblocks.

        Here is less general. It is for 2 labelblocks.

        Parameters
        ----------
        key: str
            Well position as dictionary key like "A01".

        Returns
        -------
        plt.figure
            Pointer to mpl.figure.

        Raises
        ------
        FitFirstError
            When no fitting results are yet available.

        """
        fourdigits = 1e4
        if not hasattr(self, "fittings"):
            raise FitFirstError()
        plt.style.use(["seaborn-ticks", "seaborn-whitegrid"])
        out = ["K", "sK", "SA", "sSA", "SB", "sSB"]
        out2 = ["K", "sK", "SA", "sSA", "SB", "sSB", "SA2", "sSA2", "SB2", "sSB2"]
        x = np.array(self.conc)
        xfit = np.linspace(min(x) * 0.98, max(x) * 1.02, 50)
        residues = []
        colors = []
        lines = []
        f = plt.figure(figsize=(10, 7))
        ax_data = plt.subplot2grid((3, 1), loc=(0, 0), rowspan=2)
        # labelblocks
        # for i, (lbg, df) in enumerate(zip(self.labelblocksgroups, self.fittings)):
        for i, (datafit, df) in enumerate(zip(self._datafit, self.fittings)):
            y = np.array(datafit[key]) if datafit else np.zeros_like(x)
            colors.append(plt.cm.Set2((i + 2) * 10))
            ax_data.plot(
                x, y, "o", color=colors[i], markersize=12, label="label" + str(i)
            )
            ax_data.plot(
                xfit,
                self.fz(df.K.loc[key], [df.SA.loc[key], df.SB.loc[key]], xfit),
                "-",
                lw=2,
                color=colors[i],
                alpha=0.8,
            )
            ax_data.set_xticks(ax_data.get_xticks()[1:-1])
            residues.append(
                y - self.fz(df.K.loc[key], [df.SA.loc[key], df.SB.loc[key]], x)
            )
            # Print out.
            line = [
                f"{v:.3g}" if v < fourdigits else f"{v:.0f}"
                for v in list(df[out].loc[key])
            ]
            for _i in range(4):
                line.append("")
            lines.append(line)
        # ## residues
        ax1 = plt.subplot2grid((3, 1), loc=(2, 0))
        ax1.plot(
            x, residues[0], "o-", lw=2.5, color=colors[0], alpha=0.6, markersize=12
        )
        ax2 = plt.twinx(ax1)
        ax2.plot(
            x, residues[1], "o-", lw=2.5, color=colors[1], alpha=0.6, markersize=12
        )
        plt.subplots_adjust(hspace=0)
        ax1.set_xlim(ax_data.get_xlim())
        ax_data.legend()
        # global
        fit_df = self.fittings[-1]
        lines.append(
            [
                f"{v:.3g}" if v < fourdigits else f"{v:.0f}"
                for v in list(fit_df[out2].loc[key])
            ]
        )
        ax_data.plot(
            xfit,
            self.fz(fit_df.K.loc[key], [fit_df.SA.loc[key], fit_df.SB.loc[key]], xfit),
            "b--",
            lw=0.5,
        )
        ax_data.plot(
            xfit,
            self.fz(
                fit_df.K.loc[key], [fit_df.SA2.loc[key], fit_df.SB2.loc[key]], xfit
            ),
            "b--",
            lw=0.5,
        )
        ax_data.table(cellText=lines, colLabels=out2, loc="top")
        ax1.grid(0, axis="y")  # switch off horizontal
        ax2.grid(1, axis="both")
        # ## only residues
        y = self.labelblocksgroups[0].data[key]  # type: ignore # XXX: D
        ax1.plot(
            x,
            (
                y
                - self.fz(
                    fit_df.K.loc[key], [fit_df.SA.loc[key], fit_df.SB.loc[key]], x
                )
            ),
            "--",
            lw=1.5,
            color=colors[0],
        )
        y = self.labelblocksgroups[1].data[key]  # type: ignore # XXX: D
        ax2.plot(
            x,
            (
                y
                - self.fz(
                    fit_df.K.loc[key], [fit_df.SA2.loc[key], fit_df.SB2.loc[key]], x
                )
            ),
            "--",
            lw=1.5,
            color=colors[1],
        )
        if key in self.scheme.ctrl:
            plt.title(
                "Ctrl: " + fit_df["ctrl"].loc[key] + "  [" + key + "]", {"fontsize": 16}
            )
        else:
            plt.title(key, {"fontsize": 16})
        plt.close()
        return f

    def plot_all_wells(self, path: Path) -> None:
        """Plot all wells into a pdf.

        Parameters
        ----------
        path : Path
            Where the pdf file is saved.

        Raises
        ------
        FitFirstError
            When no fitting results are yet available.

        """
        if not hasattr(self, "fittings"):
            raise FitFirstError()
        out = PdfPages(path)
        for k in self.fittings[0].loc[self.scheme.ctrl].sort_values("ctrl").index:
            out.savefig(self.plot_well(str(k)))
        for k in self.fittings[0].loc[self.keys_unk].sort_index().index:
            out.savefig(self.plot_well(str(k)))
        out.close()

    def plot_ebar(  # noqa: PLR0913
        self,
        lb: int,
        x: str = "K",
        y: str = "SA",
        xerr: str = "sK",
        yerr: str = "sSA",
        xmin: float | None = None,
        ymin: float | None = None,
        xmax: float | None = None,
        title: str | None = None,
    ) -> plt.figure:
        """Plot SA vs. K with errorbar for the whole plate."""
        if not hasattr(self, "fittings"):
            raise FitFirstError()
        fit_df = self.fittings[lb]
        with plt.style.context("fivethirtyeight"):
            f = plt.figure(figsize=(10, 10))
            if xmin:
                fit_df = fit_df[fit_df[x] > xmin]
            if xmax:
                fit_df = fit_df[fit_df[x] < xmax]
            if ymin:
                fit_df = fit_df[fit_df[y] > ymin]
            with suppress(ValueError):
                plt.errorbar(
                    fit_df[x],
                    fit_df[y],
                    xerr=fit_df[xerr],
                    yerr=fit_df[yerr],
                    fmt="o",
                    elinewidth=1,
                    markersize=10,
                    alpha=0.7,
                )
            if "ctrl" not in fit_df:
                fit_df["ctrl"] = 0
            fit_df = fit_df[~np.isnan(fit_df[x])]
            fit_df = fit_df[~np.isnan(fit_df[y])]
            for idx, xv, yv, l in zip(  # noqa: E741
                fit_df.index, fit_df[x], fit_df[y], fit_df["ctrl"]
            ):
                # x or y do not exist.# try:
                if isinstance(l, str):
                    color = "#" + hashlib.sha224(l.encode()).hexdigest()[2:8]
                    plt.text(xv, yv, l, fontsize=13, color=color)
                else:
                    plt.text(xv, yv, idx, fontsize=12)
                # x or y do not exist.# except:
                # x or y do not exist.# continue
            plt.yscale("log")
            # min(x) can be = NaN
            min_x = min(max([0.01, fit_df[x].min()]), 14)
            min_y = min(max([0.01, fit_df[y].min()]), 5000)
            plt.xlim(0.99 * min_x, 1.01 * fit_df[x].max())
            plt.ylim(0.90 * min_y, 1.10 * fit_df[y].max())
            plt.grid(1, axis="both")
            plt.ylabel(y)
            plt.xlabel(x)
            title = title if title else ""
            title += "  label:" + str(lb)
            plt.title(title, fontsize=15)
            return f

    def print_fitting(self, lb: int) -> None:
        """Print fitting parameters for the whole plate."""

        def df_print(df: pd.DataFrame) -> None:
            for i, r in df.iterrows():
                print(f"{i:s}", end=" ")
                for k in out[:2]:
                    print(f"{r[k]:7.2f}", end=" ")
                for k in out[2:]:
                    print(f"{r[k]:7.0f}", end=" ")
                print()

        fit_df = self.fittings[lb]
        if "SA2" in fit_df:
            out = ["K", "sK", "SA", "sSA", "SB", "sSB", "SA2", "sSA2", "SB2", "sSB2"]
        else:
            out = ["K", "sK", "SA", "sSA", "SB", "sSB"]
        if len(self.scheme.ctrl) > 0:
            res_ctrl = fit_df.loc[self.scheme.ctrl]
            gr = res_ctrl.groupby("ctrl")
            print("    " + " ".join([f"{x:>7s}" for x in out]))
            for g in gr:
                print(" ", g[0])
                df_print(g[1][out])
        res_unk = fit_df.loc[self.keys_unk]
        print()
        print("    " + " ".join([f"{x:>7s}" for x in out]))
        print("  UNK")
        df_print(res_unk.sort_index())

    def plot_buffer(self, title: str | None = None) -> plt.figure:
        """Plot buffers of all labelblocksgroups."""
        x = self.conc
        f, ax = plt.subplots(2, 1, figsize=(9, 9))
        for i, lbg in enumerate(self.labelblocksgroups):
            if lbg.data_buffersubtracted:
                bg = []
                bg_sd = []
                for lb in lbg.labelblocks:
                    bg.append(lb.buffer)
                    bg_sd.append(lb.buffer_sd)
            rowlabel = ["Temp"]
            lines = [
                [
                    f"{x:6.1f}"
                    for x in [
                        lb.metadata["Temperature"].value for lb in lbg.labelblocks
                    ]
                ]
            ]
            buf = {key: lbg.data[key] for key in self.scheme.buffer}  # type: ignore # XXX: D
            colors = plt.cm.Set3(np.linspace(0, 1, len(buf) + 1))
            for j, (k, v) in enumerate(buf.items(), start=1):
                rowlabel.append(k)
                lines.append([f"{x:6.1f}" for x in v])
                ax[i].plot(x, v, "o-", alpha=0.8, lw=2, markersize=3, color=colors[j])
            ax[i].errorbar(
                x,
                bg,
                yerr=bg_sd,
                fmt="o-.",
                markersize=12,
                lw=1,
                elinewidth=3,
                alpha=0.8,
                color="grey",
                label="label" + str(i),
            )
            plt.subplots_adjust(hspace=0.0)
            ax[i].legend(fontsize=16)
            if x[0] > x[-1]:  # reverse
                for line in lines:
                    line.reverse()
            ax[i].table(
                cellText=lines,
                rowLabels=rowlabel,
                loc="top",
                rowColours=colors,
                alpha=0.4,
            )
            ax[i].set_xlim(min(x) * 0.96, max(x) * 1.02)
            ax[i].set_yticks(ax[i].get_yticks()[:-1])
        ax[0].set_yticks(ax[0].get_yticks()[1:])
        ax[0].set_xticklabels("")
        f.suptitle(title, fontsize=16)
        f.tight_layout()
        return f
